import os
import shutil
import subprocess
from concurrent.futures import Future
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional

from ..common import constant
from ..common.downloader import MeDownloader
from ..common.io import MAX_CHUNK_SIZE, READ_SIZE, DecryptIO, to_decryptio
from ..common.progress_bar import (
    _progress,
    init_progress_bar,
    progress_task_exists,
)
from ..pcs import PCS_UA, BaiduPCSApi
from ..utils import human_size_to_int
from .display import display_blocked_remotepath
from .log import get_logger
from .sifter import Sifter, sift

_print = print

from rich import print
from rich.progress import TaskID

logger = get_logger(__name__)


DEFAULT_CONCURRENCY = 5
DEFAULT_CHUNK_SIZE = str(MAX_CHUNK_SIZE)


@dataclass
class DownloadParams:
    concurrency: int = DEFAULT_CONCURRENCY
    chunk_size: str = DEFAULT_CHUNK_SIZE
    quiet: bool = False
    downloader_params: List[str] = field(default_factory=list)


DEFAULT_DOWNLOADPARAMS = DownloadParams()


class Downloader(Enum):
    me = "me"
    aget_py = "aget"  # https://github.com/PeterDing/aget
    aget_rs = "ag"  # https://github.com/PeterDing/aget-rs
    aria2 = "aria2c"  # https://github.com/aria2/aria2

    # No use axel. It Can't handle URLs of length over 1024
    # axel = 'axel'  # https://github.com/axel-download-accelerator/axel

    # No use wget. the file url of baidupan only supports `Range` request

    def which(self) -> Optional[str]:
        return shutil.which(self.value)

    def download(
        self,
        url: str,
        localpath: str,
        cookies: Dict[str, Optional[str]],
        downloadparams: DownloadParams = DEFAULT_DOWNLOADPARAMS,
        out_cmd: bool = False,
        encrypt_password: bytes = b"",
    ):
        global DEFAULT_DOWNLOADER
        if not self.which():
            self = DEFAULT_DOWNLOADER

        localpath_tmp = localpath + ".tmp"

        def done_callback(fut: Future):
            err = fut.exception()
            if not err:
                shutil.move(localpath_tmp, localpath)
            else:
                logger.info("`download`: MeDownloader fails: error: %s", err)

        if self == Downloader.me:
            self._me_download(
                url,
                localpath_tmp,
                cookies=cookies,
                downloadparams=downloadparams,
                done_callback=done_callback,
                encrypt_password=encrypt_password,
            )
            return
        elif self == Downloader.aget_py:
            cmd = self._aget_py_cmd(url, localpath_tmp, cookies, downloadparams)
        elif self == Downloader.aget_rs:
            cmd = self._aget_rs_cmd(url, localpath_tmp, cookies, downloadparams)
        elif self == Downloader.aria2:
            cmd = self._aria2_cmd(url, localpath_tmp, cookies, downloadparams)
        else:
            cmd = self._aget_py_cmd(url, localpath_tmp, cookies, downloadparams)

        # Print out command
        if out_cmd:
            _print(" ".join((repr(c) for c in cmd)))
            return

        returncode = self.spawn(cmd, downloadparams.quiet)

        logger.debug("`download`: cmd returncode: %s", returncode)

        if returncode != 0:
            print(
                f"[italic]{self.value}[/italic] fails. return code: [red]{returncode}[/red]"
            )
        else:
            if encrypt_password:
                dio = to_decryptio(open(localpath_tmp, "rb"), encrypt_password)
                if isinstance(dio, DecryptIO):
                    with open(localpath, "wb") as fd:
                        while True:
                            buf = dio.read(READ_SIZE)
                            if not buf:
                                break
                            fd.write(buf)

                    os.remove(localpath_tmp)
                    return
            shutil.move(localpath_tmp, localpath)

    def spawn(self, cmd: List[str], quiet: bool = False):
        child = subprocess.run(cmd, stdout=subprocess.DEVNULL if quiet else None)
        return child.returncode

    def _me_download(
        self,
        url: str,
        localpath: str,
        cookies: Dict[str, Optional[str]],
        downloadparams: DownloadParams = DEFAULT_DOWNLOADPARAMS,
        done_callback: Optional[Callable[[Future], Any]] = None,
        encrypt_password: bytes = b"",
    ):
        headers = {
            "Cookie": f"BDUSS={cookies['BDUSS']};",
            "User-Agent": PCS_UA,
            "Connection": "Keep-Alive",
        }

        task_id: Optional[TaskID] = None
        if not downloadparams.quiet:
            init_progress_bar()
            task_id = _progress.add_task("MeDownloader", start=False, title=localpath)

        def _wrap_done_callback(fut: Future):
            if task_id is not None:
                _progress.remove_task(task_id)
            if done_callback:
                done_callback(fut)

        def monit_callback(task_id: Optional[TaskID], offset: int):
            if task_id is not None:
                _progress.update(task_id, completed=offset + 1)

        def except_callback(task_id: Optional[TaskID]):
            if task_id is not None and progress_task_exists(task_id):
                _progress.reset(task_id)

        chunk_size_int = human_size_to_int(downloadparams.chunk_size)
        meDownloader = MeDownloader(
            "GET",
            url,
            headers=headers,
            max_workers=downloadparams.concurrency,
            max_chunk_size=chunk_size_int,
            callback=monit_callback,
            encrypt_password=encrypt_password,
        )

        if task_id is not None:
            length = len(meDownloader)
            _progress.update(task_id, total=length)
            _progress.start_task(task_id)

        meDownloader.download(
            Path(localpath),
            task_id=task_id,
            continue_=True,
            done_callback=_wrap_done_callback,
            except_callback=except_callback,
        )

    def _aget_py_cmd(
        self,
        url: str,
        localpath: str,
        cookies: Dict[str, Optional[str]],
        downloadparams: DownloadParams = DEFAULT_DOWNLOADPARAMS,
    ):
        _ck = f"Cookie: BDUSS={cookies['BDUSS']};"

        # This is an error of aget-py
        chunk_size = human_size_to_int(downloadparams.chunk_size)
        if chunk_size == MAX_CHUNK_SIZE:
            chunk_size -= constant.OneM

        cmd = [
            self.which(),
            url,
            "-o",
            localpath,
            "-H",
            f"User-Agent: {PCS_UA}",
            "-H",
            "Connection: Keep-Alive",
            "-H",
            _ck,
            "-s",
            str(downloadparams.concurrency),
            "-k",
            str(chunk_size),
            *downloadparams.downloader_params,
        ]
        return cmd

    def _aget_rs_cmd(
        self,
        url: str,
        localpath: str,
        cookies: Dict[str, Optional[str]],
        downloadparams: DownloadParams = DEFAULT_DOWNLOADPARAMS,
    ):
        _ck = f"Cookie: BDUSS={cookies['BDUSS']};"
        cmd = [
            self.which(),
            url,
            "-o",
            localpath,
            "-H",
            f"User-Agent: {PCS_UA}",
            "-H",
            "Connection: Keep-Alive",
            "-H",
            _ck,
            "-s",
            str(downloadparams.concurrency),
            "-k",
            downloadparams.chunk_size,
            *downloadparams.downloader_params,
        ]
        return cmd

    def _aria2_cmd(
        self,
        url: str,
        localpath: str,
        cookies: Dict[str, Optional[str]],
        downloadparams: DownloadParams = DEFAULT_DOWNLOADPARAMS,
    ):
        _ck = f"Cookie: BDUSS={cookies['BDUSS']};"
        directory, filename = os.path.split(localpath)
        cmd = [
            self.which(),
            url,
            "-c",
            "--dir",
            directory,
            "-o",
            filename,
            "--header",
            f"User-Agent: {PCS_UA}",
            "--header",
            "Connection: Keep-Alive",
            "--header",
            _ck,
            "-s",
            str(downloadparams.concurrency),
            "-k",
            downloadparams.chunk_size,
            *downloadparams.downloader_params,
        ]
        return cmd


DEFAULT_DOWNLOADER = Downloader.me


def download_file(
    api: BaiduPCSApi,
    remotepath: str,
    localdir: str,
    downloader: Downloader = DEFAULT_DOWNLOADER,
    downloadparams: DownloadParams = DEFAULT_DOWNLOADPARAMS,
    out_cmd: bool = False,
    encrypt_password: bytes = b"",
):
    localpath = Path(localdir) / os.path.basename(remotepath)

    # Make sure parent directory existed
    if not localpath.parent.exists():
        localpath.parent.mkdir(parents=True)

    if not out_cmd and localpath.exists():
        print(f"[yellow]{localpath}[/yellow] is ready existed.")
        return

    dlink = api.download_link(remotepath)
    if not dlink:
        display_blocked_remotepath(remotepath)
        return

    if downloader != Downloader.me:
        print(f"[italic blue]Download[/italic blue]: {remotepath} to {localpath}")
    downloader.download(
        dlink,
        str(localpath),
        api.cookies,
        downloadparams=downloadparams,
        out_cmd=out_cmd,
        encrypt_password=encrypt_password,
    )


def download_dir(
    api: BaiduPCSApi,
    remotedir: str,
    localdir: str,
    sifters: List[Sifter] = [],
    recursive: bool = False,
    from_index: int = 0,
    downloader: Downloader = DEFAULT_DOWNLOADER,
    downloadparams=DEFAULT_DOWNLOADPARAMS,
    out_cmd: bool = False,
    encrypt_password: bytes = b"",
):
    remotepaths = api.list(remotedir)
    remotepaths = sift(remotepaths, sifters, recursive=recursive)
    for rp in remotepaths[from_index:]:
        if rp.is_file:
            download_file(
                api,
                rp.path,
                localdir,
                downloader,
                downloadparams=downloadparams,
                out_cmd=out_cmd,
                encrypt_password=encrypt_password,
            )
        else:  # is_dir
            if recursive:
                _localdir = Path(localdir) / os.path.basename(rp.path)
                download_dir(
                    api,
                    rp.path,
                    str(_localdir),
                    sifters=sifters,
                    recursive=recursive,
                    from_index=from_index,
                    downloader=downloader,
                    downloadparams=downloadparams,
                    out_cmd=out_cmd,
                    encrypt_password=encrypt_password,
                )


def download(
    api: BaiduPCSApi,
    remotepaths: List[str],
    localdir: str,
    sifters: List[Sifter] = [],
    recursive: bool = False,
    from_index: int = 0,
    downloader: Downloader = DEFAULT_DOWNLOADER,
    downloadparams: DownloadParams = DEFAULT_DOWNLOADPARAMS,
    out_cmd: bool = False,
    encrypt_password: bytes = b"",
):
    """Download `remotepaths` to the `localdir`

    Args:
        `from_index` (int): The start index of downloading entries from EACH remote directory
    """

    logger.debug(
        "`download`: sifters: %s, recursive: %s, from_index: %s, "
        "downloader: %s, downloadparams: %s, out_cmd: %s, has encrypt_password: %s",
        sifters,
        recursive,
        from_index,
        downloader,
        downloadparams,
        out_cmd,
        bool(encrypt_password),
    )
    logger.debug(
        "`download`: remotepaths should be uniq %s == %s",
        len(remotepaths),
        len(set(remotepaths)),
    )

    # assert (
    #     human_size_to_int(downloadparams.chunk_size) <= MAX_CHUNK_SIZE
    # ), f"`chunk_size` must be less or equal then {human_size(MAX_CHUNK_SIZE)}"

    for rp in remotepaths:
        if not api.exists(rp):
            print(f"[yellow]WARNING[/yellow]: `{rp}` does not exist.")
            continue

        if api.is_file(rp):
            download_file(
                api,
                rp,
                localdir,
                downloader=downloader,
                downloadparams=downloadparams,
                out_cmd=out_cmd,
                encrypt_password=encrypt_password,
            )
        else:
            _localdir = str(Path(localdir) / os.path.basename(rp))
            download_dir(
                api,
                rp,
                _localdir,
                sifters=sifters,
                recursive=recursive,
                from_index=from_index,
                downloader=downloader,
                downloadparams=downloadparams,
                out_cmd=out_cmd,
                encrypt_password=encrypt_password,
            )

    if downloader == Downloader.me:
        MeDownloader._exit_executor()

    _progress.stop()



from pathlib import Path

# Local data directory
DATA_DIR = Path("~") / ".baidupcs-py"

# Account data path
ACCOUNT_DATA_PATH = DATA_DIR / "accounts.pk"

# Localstorage data path
RAPIDUPLOADINFO_PATH = DATA_DIR / "rapiduploadinfo.sqlite3"

# Logging path
LOG_PATH = DATA_DIR / "running.log"

# Logging level
LOG_LEVEL = "CRITICAL"



from .api import BaiduPCSApi, BaiduPCS
from .pcs import PCS_UA
from .errors import BaiduPCSError
from .inner import (
    PcsRapidUploadInfo,
    PcsFile,
    PcsMagnetFile,
    PcsSharedLink,
    PcsSharedPath,
    PcsQuota,
    PcsAuth,
    PcsUserProduct,
    PcsUser,
    CloudTask,
    FromTo,
)


__all__ = [
    "BaiduPCS",
    "BaiduPCSApi",
    "BaiduPCSError",
    "PcsRapidUploadInfo",
    "PcsFile",
    "PcsMagnetFile",
    "PcsSharedLink",
    "PcsSharedPath",
    "PcsQuota",
    "PcsAuth",
    "PcsUserProduct",
    "PcsUser",
    "CloudTask",
    "FromTo",
    "PCS_UA",
]



from typing import Iterator
from pathlib import Path
import os
from os import PathLike

from .platform import IS_WIN


def exists(localpath: PathLike) -> bool:
    localpath = Path(localpath)
    return localpath.exists()


def is_file(localpath: PathLike) -> bool:
    localpath = Path(localpath)
    return localpath.is_file()


def is_dir(localpath: PathLike) -> bool:
    localpath = Path(localpath)
    return localpath.is_dir()


def walk(localpath: PathLike) -> Iterator[str]:
    for root, _, files in os.walk(Path(localpath)):
        r = Path(root)
        for fl in files:
            yield (r / fl).as_posix()


def join_path(source: PathLike, dest: PathLike) -> str:
    """Join posix paths"""

    _path = (Path(source) / dest).as_posix()
    has_root = _path.startswith("/")
    if not has_root:
        _path = "/" + _path

    path = Path(_path).resolve().as_posix()

    if IS_WIN:
        p = path.split(":", 1)[-1]
        if not has_root:
            return p[1:]
        else:
            return p
    else:
        if not has_root:
            return path[1:]
        else:
            return path



from ..pcs import BaiduPCSApi, BaiduPCSError
from .display import display_files, display_from_to


def makedir(api: BaiduPCSApi, *remotedirs: str, show: bool = False):
    pcs_files = []
    for d in remotedirs:
        pcs_file = api.makedir(d)
        pcs_files.append(pcs_file)

    if show:
        display_files(pcs_files, "/", show_absolute_path=True)


def move(api: BaiduPCSApi, *remotepaths: str, show: bool = False):
    from_to_list = api.move(*remotepaths)

    if show:
        display_from_to(*from_to_list)


def rename(api: BaiduPCSApi, source: str, dest: str, show: bool = False):
    from_to_list = api.rename(source, dest)

    if show:
        display_from_to(from_to_list)


def copy(api: BaiduPCSApi, *remotepaths: str, show: bool = False):
    from_to_list = api.copy(*remotepaths)

    if show:
        display_from_to(*from_to_list)


def remove(api: BaiduPCSApi, *remotepaths: str):
    try:
        api.remove(*remotepaths)
    except BaiduPCSError as err:
        if err.error_code != 31066:
            raise err



from os import PathLike
from pathlib import Path

from ..common.io import READ_SIZE, to_decryptio
from ..common.path import exists


def decrypt_file(
    from_encrypted: PathLike, to_decrypted: PathLike, encrypt_password: bytes = b""
):
    assert exists(from_encrypted)

    dio = to_decryptio(open(from_encrypted, "rb"), encrypt_password=encrypt_password)

    dpath = Path(to_decrypted)
    dir_ = dpath.parent
    if not dir_.exists():
        dir_.mkdir(parents=True)

    with dpath.open("wb") as dfd:
        while True:
            data = dio.read(READ_SIZE)
            if not data:
                break
            dfd.write(data)



from base64 import b64decode
from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path
from threading import Semaphore
from typing import Any, Dict, List, Optional

from rich import print

from ..commands.display import (
    display_rapid_upload_infos,
    display_rapid_upload_links,
)
from ..common.concurrent import sure_release
from ..common.constant import CPU_NUM
from ..common.localstorage import RapidUploadInfo, save_rapid_upload_info
from ..common.path import join_path
from ..pcs import BaiduPCSApi
from ..pcs.inner import PcsRapidUploadInfo


def _display(
    rows: List[Dict[str, Any]],
    hash_link_protocol: str = PcsRapidUploadInfo.default_hash_link_protocol(),
    show_all: bool = False,
    only_hash_link: bool = False,
):
    if show_all and not only_hash_link:
        display_rapid_upload_infos(rows)
    else:
        display_rapid_upload_links(
            rows, hash_link_protocol=hash_link_protocol, only_hash_link=only_hash_link
        )


def rapid_upload_list(
    rapiduploadinfo_file: str,
    ids: List[int] = [],
    filename: bool = False,
    time: bool = False,
    size: bool = False,
    localpath: bool = False,
    remotepath: bool = False,
    user_id: bool = False,
    user_name: bool = False,
    desc: bool = False,
    limit: int = 0,
    offset: int = -1,
    hash_link_protocol: str = PcsRapidUploadInfo.default_hash_link_protocol(),
    show_all: bool = False,
    only_hash_link: bool = False,
):
    rapiduploadinfo = RapidUploadInfo(rapiduploadinfo_file)
    rows = rapiduploadinfo.list(
        ids=ids,
        by_filename=filename,
        by_time=time,
        by_size=size,
        by_localpath=localpath,
        by_remotepath=remotepath,
        by_user_id=user_id,
        by_user_name=user_name,
        desc=desc,
        limit=limit,
        offset=offset,
    )
    _display(
        rows,
        hash_link_protocol=hash_link_protocol,
        show_all=show_all,
        only_hash_link=only_hash_link,
    )


def rapid_upload_search(
    rapiduploadinfo_file: str,
    keyword: str,
    in_filename: bool = False,
    in_localpath: bool = False,
    in_remotepath: bool = False,
    in_user_name: bool = False,
    in_md5: bool = False,
    hash_link_protocol: str = PcsRapidUploadInfo.default_hash_link_protocol(),
    show_all: bool = False,
    only_hash_link: bool = False,
):
    rapiduploadinfo = RapidUploadInfo(rapiduploadinfo_file)
    rows = rapiduploadinfo.search(
        keyword,
        in_filename=in_filename,
        in_localpath=in_localpath,
        in_remotepath=in_remotepath,
        in_user_name=in_user_name,
        in_md5=in_md5,
    )
    _display(
        rows,
        hash_link_protocol=hash_link_protocol,
        show_all=show_all,
        only_hash_link=only_hash_link,
    )


def rapid_upload_delete(rapiduploadinfo_file: str, ids: List[int]):
    rapiduploadinfo = RapidUploadInfo(rapiduploadinfo_file)
    for id in ids:
        rapiduploadinfo.delete(id)


def _parse_link(link: str) -> Any:
    """Parse rapid upload link

    Format 1: cs3l://<content_md5>#<slice_md5>#<content_crc3>#<content_length>#<filename>
    Format 2: <content_md5>#<slice_md5>#<content_length>#<filename>
    Format 3: bdpan://{base64(<filename>|<content_length>|<content_md5>|<slice_md5>)}

    Format 2 is from https://blog.jixun.moe/du-code-gen , using at
        https://greasyfork.org/zh-CN/scripts/397324-%E7%A7%92%E4%BC%A0%E9%93%BE%E6%8E%A5%E6%8F%90%E5%8F%96
    """

    if link.startswith("cs3l://"):
        chunks = link[7:].split("#", 4)
        assert len(chunks) == 5
    elif link.startswith("bdpan://"):
        link = b64decode(link[8:]).decode("utf-8")
        link = link[::-1]
        chunks = link.split("|", 3)

        assert len(chunks) == 4

        chunks = [c[::-1] for c in chunks]
        chunks = [chunks[1], chunks[0], chunks[2], chunks[3]]
    else:
        chunks = link.split("#", 3)
        assert len(chunks) == 4

    if len(chunks) == 4:
        content_crc32 = "0"
        content_md5, slice_md5, content_length, filename = chunks
    elif len(chunks) == 5:
        content_md5, slice_md5, content_crc32, content_length, filename = chunks
        content_crc32 = content_crc32 or "0"  # content_crc32 can be ''
    else:
        raise RuntimeError("Here is Unreachable")

    assert len(content_md5) == len(slice_md5) == 32

    filename = filename.replace("%20", " ")
    return (slice_md5, content_md5, int(content_crc32), int(content_length), filename)


def rapid_upload(
    api: BaiduPCSApi,
    remotedir: str,
    link: str = "",
    slice_md5: str = "",
    content_md5: str = "",
    content_crc32: int = 0,
    content_length: int = 0,
    filename: str = "",
    no_ignore_existing: bool = False,
    rapiduploadinfo_file: Optional[str] = None,
    user_id: Optional[int] = None,
    user_name: Optional[str] = None,
):
    """Rapid upload with params

    If given `link` and `filename`, then filename of link will be replace by `filename`
    """

    if link:
        slice_md5, content_md5, content_crc32, content_length, _filename = _parse_link(
            link
        )
        content_crc32 = content_crc32 or 0
        filename = filename or _filename

    remotepath = join_path(Path(remotedir), Path(filename))

    assert all(
        [slice_md5, content_md5, content_length]
    ), f"`rapid_upload`: parsing rapid upload link fails: {link}"

    if not no_ignore_existing:
        if api.exists(remotepath):
            return

    try:
        pcs_file = api.rapid_upload_file(
            slice_md5,
            content_md5,
            content_crc32,
            content_length,
            remotepath,
            ondup="overwrite",
        )
        if rapiduploadinfo_file:
            save_rapid_upload_info(
                rapiduploadinfo_file,
                slice_md5,
                content_md5,
                content_crc32,
                content_length,
                remotepath=remotepath,
                user_id=user_id,
                user_name=user_name,
            )
        print(f"[i blue]Save to[/i blue] {pcs_file.path}")
    except Exception as err:
        link = PcsRapidUploadInfo(
            slice_md5=slice_md5,
            content_md5=content_md5,
            content_crc32=content_crc32,
            content_length=content_length,
            remotepath=remotepath,
        ).cs3l()
        print(
            f"[i yellow]Rapid Upload fails[/i yellow]: error: {err} link: {link}",
        )


def rapid_upload_links(
    api: BaiduPCSApi,
    remotedir: str,
    links: List[str] = [],
    no_ignore_existing: bool = False,
    rapiduploadinfo_file: Optional[str] = None,
    user_id: Optional[int] = None,
    user_name: Optional[str] = None,
    max_workers: int = CPU_NUM,
):
    """Rapid Upload multi links"""

    semaphore = Semaphore(max_workers)
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        futs = []
        for link in links:
            link = link.strip()
            if not link:
                continue

            semaphore.acquire()
            fut = executor.submit(
                sure_release,
                semaphore,
                rapid_upload,
                api,
                remotedir,
                link=link,
                no_ignore_existing=no_ignore_existing,
                rapiduploadinfo_file=rapiduploadinfo_file,
                user_id=user_id,
                user_name=user_name,
            )
            futs.append(fut)

        for fut in as_completed(futs):
            pass



import sys

if sys.version_info < (3, 9):
    from typing_extensions import Annotated
else:
    from typing import Annotated

from typing import Optional, Dict, Any
from pathlib import Path
import os
import mimetypes
import secrets
import copy
from urllib.parse import quote

import uvicorn

from ..pcs import BaiduPCSApi
from ..common.io import RangeRequestIO
from ..common.constant import CPU_NUM
from ..common.path import join_path
from ..utils import format_date

from fastapi import Depends, FastAPI, Request, status, HTTPException
from fastapi.security import HTTPBasic, HTTPBasicCredentials
from fastapi.responses import HTMLResponse, StreamingResponse, Response
from jinja2 import Template

from rich import print

mimetypes.init()

app = FastAPI()

_api: Optional[BaiduPCSApi] = None
_root_dir: str = "/"
_encrypt_password: bytes = b""

# For Auth
_username: Optional[str] = None
_password: Optional[str] = None

# This template is from https://github.com/rclone/rclone/blob/master/cmd/serve/httplib/serve/data/templates/index.html
_html_tempt: Template = Template(
    (Path(__file__).parent / "index.html").open(encoding="utf-8").read()
)


def fake_io(io: RangeRequestIO, start: int = 0, end: int = -1):
    for b in io._auto_decrypt_request.read((start, end)):
        yield b


async def handle_request(
    request: Request,
    remotepath: str,
    order: str = "asc",  # desc , asc
    sort: str = "name",  # name, time, size
):
    desc = order == "desc"
    name = sort == "name"
    time = sort == "time"
    size = sort == "size"

    global _root_dir
    global _api
    assert _api

    remotepath = remotepath.strip("/")

    _rp = join_path(Path(_root_dir), Path(remotepath))

    # Anti path traversal attack
    if not _rp.startswith(_root_dir):
        raise HTTPException(status_code=404, detail="Item not found")

    _range = request.headers.get("range")

    if not _api.exists(_rp):
        raise HTTPException(status_code=404, detail="Item not found")

    is_dir = _api.is_dir(_rp)
    if is_dir:
        chunks = ["/"] + (remotepath.split("/") if remotepath != "" else [])
        navigation = [
            (i - 1, "../" * (len(chunks) - i), name) for i, name in enumerate(chunks, 1)
        ]
        pcs_files = _api.list(_rp, desc=desc, name=name, time=time, size=size)
        entries = []
        for f in pcs_files:
            p = Path(f.path)
            entries.append(
                (
                    f.is_dir,
                    p.name,
                    quote(p.name),
                    f.size,
                    format_date(f.local_mtime or 0),
                )
            )
        cn = _html_tempt.render(
            root_dir=remotepath, navigation=navigation, entries=entries
        )
        return HTMLResponse(cn)
    else:
        try:
            fs = _api.file_stream(_rp, encrypt_password=_encrypt_password)
        except Exception as err:
            print("Error:", err)
            raise HTTPException(
                status_code=500, detail=f"Error: {err}, remotepath: {_rp}"
            )

        if not fs:
            raise HTTPException(status_code=404, detail=f"No download link: {_rp}")

        length = len(fs)

        headers: Dict[str, str] = {
            "accept-ranges": "bytes",
            "connection": "Keep-Alive",
            "access-control-allow-origin": "*",
        }

        ext = os.path.splitext(remotepath)[-1]
        content_type = mimetypes.types_map.get(ext)

        if content_type:
            headers["content-type"] = content_type

        if _range and fs.seekable():
            assert _range.startswith("bytes=")

            status_code = 206
            start, end = _range[6:].split("-")
            _s, _e = int(start or 0), int(end or length - 1) + 1
            _iter_io = fake_io(fs, _s, _e)
            headers["content-range"] = f"bytes {_s}-{_e-1}/{length}"
            headers["content-length"] = str(_e - _s)
        else:
            status_code = 200
            _iter_io = fake_io(fs)
            headers["content-length"] = str(length)

        return StreamingResponse(_iter_io, status_code=status_code, headers=headers)


_security = HTTPBasic()


def to_auth(credentials: HTTPBasicCredentials = Depends(_security)) -> str:
    correct_username = secrets.compare_digest(credentials.username, _username or "")
    correct_password = secrets.compare_digest(credentials.password, _password or "")
    if not (correct_username and correct_password):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect email or password",
            headers={"WWW-Authenticate": "Basic"},
        )
    return credentials.username


def make_auth_http_server(path: str = ""):
    @app.get("%s/{remotepath:path}" % path)
    async def auth_http_server(
        username: Annotated[str, Depends(to_auth)],
        response: Annotated[Any, Depends(handle_request)],
    ):
        # async def auth_http_server(username: str = Depends(to_auth), response: Response = Depends(handle_request)):
        if username:
            return response


def make_http_server(path: str = ""):
    @app.get("%s/{remotepath:path}" % path)
    async def http_server(response: Annotated[Any, Depends(handle_request)]):
        return response


def start_server(
    api: BaiduPCSApi,
    root_dir: str = "/",
    path: str = "",
    host: str = "localhost",
    port: int = 8000,
    workers: int = CPU_NUM,
    encrypt_password: bytes = b"",
    log_level: str = "info",
    username: Optional[str] = None,
    password: Optional[str] = None,
):
    """Create a http server on remote `root_dir`"""

    global _encrypt_password
    _encrypt_password = encrypt_password

    global _root_dir
    _root_dir = root_dir

    global _api
    if not _api:
        _api = api

    global _username
    if not _username:
        _username = username

    global _password
    if not _password:
        _password = password

    if path == "/" or not path:
        path = ""
    else:
        path = "/" + path.strip("/")

    print(f"[yellow]Server running on[/yellow] [b]http://{host}:{port}{path}/[/b]")

    if _username and _password:
        make_auth_http_server(path)
    else:
        make_http_server(path)

    log_config = copy.deepcopy(uvicorn.config.LOGGING_CONFIG)
    log_config["formatters"]["access"]["fmt"] = (
        '%(asctime)s - %(levelprefix)s %(client_addr)s - "%(request_line)s" %(status_code)s'
    )
    uvicorn.run(
        "baidupcs_py.commands.server:app",
        host=host,
        port=port,
        log_level=log_level,
        log_config=log_config,
        workers=1,
    )



import pickle
from os import PathLike
from pathlib import Path
from typing import Dict, List, NamedTuple, Optional

from rich import print

from baidu.baidupcs import BaiduPCSApi, PcsUser
from baidu.commands.env import ACCOUNT_DATA_PATH
from baidu.common.path import join_path


class Account(NamedTuple):
    user: PcsUser

    # current working directory
    pwd: str = "/"
    encrypt_password: Optional[str] = None

    # In version 2, salt is generated by `os.urandom`.
    # Here it useless.
    salt: Optional[str] = None

    # Account Name which can be set by hand
    account_name: str = ""

    def pcsapi(self) -> BaiduPCSApi:
        auth = self.user.auth
        assert auth, f"{self}.user.auth is None"
        return BaiduPCSApi(
            bduss=auth.bduss,
            stoken=auth.stoken,
            ptoken=auth.ptoken,
            cookies=auth.cookies,
            user_id=self.user.user_id,
        )

    @staticmethod
    def from_bduss(
        bduss: str,
        cookies: Dict[str, Optional[str]] = {},
        account_name: str = "",
    ) -> "Account":
        api = BaiduPCSApi(bduss=bduss, cookies=cookies)
        user = api.user_info()
        return Account(user=user, account_name=account_name or user.user_name or "")


class AccountManager:
    """Account Manager

    Manage all accounts
    """

    def __init__(self, data_path: Optional[PathLike] = None):
        self._accounts: Dict[int, Account] = {}  # user_id (int) -> Account
        self._who: Optional[int] = None  # user_id (int)
        self._data_path = data_path

    @staticmethod
    def load_data(data_path: PathLike) -> "AccountManager":
        try:
            data_path = Path(data_path).expanduser()
            am = pickle.load(data_path.open("rb"))
            _compat_account_manager(am)
            return am
        except Exception:
            return AccountManager(data_path=data_path)

    @property
    def accounts(self) -> List[Account]:
        """All accounts"""

        return list(self._accounts.values())

    def set_account_name(self, account_name: str, user_id: int = None):
        """Set account name"""

        user_id = user_id or self._who

        assert user_id, "No recent user"

        account = self._accounts.get(user_id)

        assert account

        account = account._replace(account_name=account_name)
        self._accounts[user_id] = account

    def set_encrypt_password(
        self, encrypt_password: Optional[str] = None, salt: Optional[str] = None
    ):
        """Set encryption key"""

        assert self._who, "No recent user"

        account = self._accounts.get(self._who)

        assert account

        account = account._replace(encrypt_password=encrypt_password, salt=salt)
        self._accounts[self._who] = account

    def cd(self, remotedir: str = "/"):
        """Change current working directory"""

        assert self._who, "No recent user"

        account = self._accounts.get(self._who)

        assert account

        pwd = join_path(Path(account.pwd), Path(remotedir))
        account = account._replace(pwd=pwd)
        self._accounts[self._who] = account

    @property
    def pwd(self) -> str:
        """Current working directory of recent user"""

        assert self._who, "No recent user"

        account = self._accounts.get(self._who)

        assert account

        return account.pwd

    def who(self, user_id: Optional[int] = None) -> Optional[Account]:
        """Return recent `Account`"""

        user_id = user_id or self._who
        if user_id:
            return self._accounts.get(user_id)
        else:
            return None

    def update(self, user_id: Optional[int] = None):
        """Update user_info"""

        user_id = user_id or self._who
        if user_id:
            account = self._accounts.get(user_id)
            if not account:
                return None

            api = account.pcsapi()
            user = api.user_info()

            self._accounts[user_id] = account._replace(user=user)

    def su(self, user_id: int):
        """Change recent user with `PcsUser.user_id`

        Args:
            who (int): `PcsUser.user_id`
        """

        assert user_id in self._accounts, f"No user {user_id}"

        self._who = user_id

    def add_account(self, account: Account):
        """Add an account to the manager"""

        self._accounts[account.user.user_id] = account

    def delete_account(self, user_id: int):
        """Delete a user

        Args:
            who (int): `PcsUser.user_id`
        """

        if user_id in self._accounts:
            del self._accounts[user_id]
        if user_id == self._who:
            self._who = None

    def save(self, data_path: Optional[PathLike] = None):
        """Serialize to local path"""

        data_path = data_path or self._data_path
        assert data_path, "No data path"

        data_path = Path(data_path).expanduser()
        if not data_path.parent.exists():
            data_path.parent.mkdir(parents=True)

        pickle.dump(self, open(data_path, "wb"))


def _compat_account_manager(am: AccountManager):
    """Update stored account manager to compat current version"""
    _compat_v0_5_9(am)
    _set_account_names(am)


def _compat_v0_5_9(am: AccountManager):
    for user_id, account in am._accounts.items():
        old_products = account.user.products
        if isinstance(old_products, dict):  # type: ignore
            print(  # type: ignore
                "[i yellow]Update[/i yellow]: Transfer PcsUser format for ^v0.5.9: "
                f"user_id: {user_id}"
            )
            am.update(user_id)

    data_path = am._data_path
    if not data_path or not Path(data_path).exists():
        am._data_path = ACCOUNT_DATA_PATH

    am.save()


def _set_account_names(am: AccountManager):
    """Set account's name with its user's name if that is empty"""

    for user_id, account in am._accounts.items():
        if not account.account_name:
            am._accounts[user_id] = account._replace(
                account_name=account.user.user_name or ""
            )



import os
import shutil
import subprocess
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional
from urllib.parse import quote

from ..pcs import PCS_UA, BaiduPCSApi
from .display import display_blocked_remotepath
from .errors import CommandError
from .sifter import Sifter, sift
from ..common import random
from ..common.file_type import MEDIA_EXTS

_print = print

from rich import print


def _with_media_ext(path: str) -> bool:
    ext = os.path.splitext(path)[-1].lower()
    if ext in MEDIA_EXTS:
        return True
    else:
        return False


class Player(Enum):
    mpv = "mpv"  # https://mpv.io

    def which(self) -> Optional[str]:
        return shutil.which(self.value)

    def play(
        self,
        url: str,
        cookies: Dict[str, Optional[str]],
        m3u8: bool = False,
        quiet: bool = False,
        player_params: List[str] = [],
        out_cmd: bool = False,
        use_local_server: bool = False,
    ):
        global DEFAULT_PLAYER
        if not self.which():
            print(
                f"[yellow]No player {self.name}[/yellow], using default player: {DEFAULT_PLAYER.name}"
            )
            self = DEFAULT_PLAYER
        if not self.which():
            raise CommandError(f"No player: {self.name}")

        if self == Player.mpv:
            cmd = self._mpv_cmd(
                url,
                cookies,
                m3u8=m3u8,
                quiet=quiet,
                player_params=player_params,
                use_local_server=use_local_server,
            )
        else:
            cmd = self._mpv_cmd(
                url,
                cookies,
                m3u8=m3u8,
                quiet=quiet,
                player_params=player_params,
                use_local_server=use_local_server,
            )

        # Print out command
        if out_cmd:
            _print(" ".join((repr(c) for c in cmd)))
            return

        returncode = self.spawn(cmd)
        if returncode != 0:
            print(
                f"[italic]{self.value}[/italic] fails. return code: [red]{returncode}[/red]"
            )

    def spawn(
        self,
        cmd: List[str],
        quiet: bool = False,
        player_params: List[str] = [],
    ):
        child = subprocess.run(
            cmd,
            stdout=subprocess.DEVNULL if quiet else None,
        )
        return child.returncode

    def _mpv_cmd(
        self,
        url: str,
        cookies: Dict[str, Optional[str]],
        m3u8: bool = False,
        quiet: bool = False,
        player_params: List[str] = [],
        use_local_server: bool = False,
    ):
        if use_local_server:
            cmd = [self.which(), url, *player_params]
        else:
            _ck = "Cookie: " + "; ".join(
                [f"{k}={v if v is not None else ''}" for k, v in cookies.items()]
            )
            cmd = [
                self.which(),
                url,
                "--no-ytdl",
                "--http-header-fields="
                f'"User-Agent: {PCS_UA}","{_ck}","Connection: Keep-Alive"',
                *player_params,
            ]
        if not use_local_server and m3u8:
            cmd.append(
                "--stream-lavf-o-append="
                "protocol_whitelist=file,http,https,tcp,tls,crypto,hls,applehttp"
            )
        if quiet:
            cmd.append("--really-quiet")
        return cmd


DEFAULT_PLAYER = Player.mpv
DEFAULT_TEMP_M3U8 = str(Path("~").expanduser() / ".baidupcs-py" / "recent.m3u8")


def play_file(
    api: BaiduPCSApi,
    remotepath: str,
    player: Player = DEFAULT_PLAYER,
    player_params: List[str] = [],
    m3u8: bool = False,
    quiet: bool = False,
    ignore_ext: bool = False,
    out_cmd: bool = False,
    local_server: str = "",
):
    if not ignore_ext and not _with_media_ext(remotepath):
        return

    print(f"[italic blue]Play[/italic blue]: {remotepath} {'(m3u8)' if m3u8 else ''}")

    # For typing
    url: Optional[str] = None

    if m3u8:
        m3u8_cn = api.m3u8_stream(remotepath)
        with open(DEFAULT_TEMP_M3U8, "w") as fd:
            fd.write(m3u8_cn)
        url = DEFAULT_TEMP_M3U8

    use_local_server = bool(local_server)

    if use_local_server:
        url = f"{local_server}{quote(remotepath)}"
        print("url:", url)
    else:
        url = api.download_link(remotepath)
        if not url:
            display_blocked_remotepath(remotepath)
            return

    player.play(
        url,
        api.cookies,
        m3u8=m3u8,
        quiet=quiet,
        player_params=player_params,
        out_cmd=out_cmd,
        use_local_server=use_local_server,
    )


def play_dir(
    api: BaiduPCSApi,
    remotedir: str,
    sifters: List[Sifter] = [],
    recursive: bool = False,
    from_index: int = 0,
    player: Player = DEFAULT_PLAYER,
    player_params: List[str] = [],
    m3u8: bool = False,
    quiet: bool = False,
    shuffle: bool = False,
    ignore_ext: bool = False,
    out_cmd: bool = False,
    local_server: str = "",
):
    remotepaths = api.list(remotedir)
    remotepaths = sift(remotepaths, sifters, recursive=recursive)

    if shuffle:
        random.shuffle(remotepaths)

    for rp in remotepaths[from_index:]:
        if rp.is_file:
            play_file(
                api,
                rp.path,
                player,
                player_params=player_params,
                m3u8=m3u8,
                quiet=quiet,
                ignore_ext=ignore_ext,
                out_cmd=out_cmd,
                local_server=local_server,
            )
        else:  # is_dir
            if recursive:
                play_dir(
                    api,
                    rp.path,
                    sifters=sifters,
                    recursive=recursive,
                    from_index=from_index,
                    player=player,
                    player_params=player_params,
                    m3u8=m3u8,
                    quiet=quiet,
                    shuffle=shuffle,
                    ignore_ext=ignore_ext,
                    out_cmd=out_cmd,
                    local_server=local_server,
                )


def play(
    api: BaiduPCSApi,
    remotepaths: List[str],
    sifters: List[Sifter] = [],
    recursive: bool = False,
    from_index: int = 0,
    player: Player = DEFAULT_PLAYER,
    player_params: List[str] = [],
    m3u8: bool = False,
    quiet: bool = False,
    shuffle: bool = False,
    ignore_ext: bool = False,
    out_cmd: bool = False,
    local_server: str = "",
):
    """Play media file in `remotepaths`

    Args:
        `from_index` (int): The start index of playing entries from EACH remote directory
    """

    if shuffle:
        random.shuffle(remotepaths)

    for rp in remotepaths:
        if not api.exists(rp):
            print(f"[yellow]WARNING[/yellow]: `{rp}` does not exist.")
            continue

        if api.is_file(rp):
            play_file(
                api,
                rp,
                player=player,
                player_params=player_params,
                m3u8=m3u8,
                quiet=quiet,
                ignore_ext=ignore_ext,
                out_cmd=out_cmd,
                local_server=local_server,
            )
        else:
            play_dir(
                api,
                rp,
                sifters=sifters,
                recursive=recursive,
                from_index=from_index,
                player=player,
                player_params=player_params,
                m3u8=m3u8,
                quiet=quiet,
                shuffle=shuffle,
                ignore_ext=ignore_ext,
                out_cmd=out_cmd,
                local_server=local_server,
            )



import json
import random
import re
import time
import urllib
from base64 import standard_b64encode
from enum import Enum
from pathlib import Path
from typing import IO, Any, Callable, Dict, List, Optional, Union
from urllib.error import HTTPError
from urllib.parse import quote_plus, urlparse

import requests  # type: ignore
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from typing_extensions import Literal

from ..common.cache import timeout_cache
from ..common.crypto import calu_md5, calu_sha1
from ..common.date import now_timestamp
from ..common.io import MAX_CHUNK_SIZE, RangeRequestIO
from ..utils import dump_json
from .errors import BaiduPCSError, assert_ok
from .phone import get_phone_model, sum_IMEI

PCS_BAIDU_COM = "https://pcs.baidu.com"
# PCS_BAIDU_COM = 'http://127.0.0.1:8888'
PAN_BAIDU_COM = "https://pan.baidu.com"
# PAN_BAIDU_COM = 'http://127.0.0.1:8888'

# PCS_UA = "netdisk;P2SP;2.2.90.43;WindowsBaiduYunGuanJia;netdisk;11.4.5;android-android;11.0;JSbridge4.4.0;LogStatistic"
# PCS_UA = "netdisk;P2SP;2.2.91.26;netdisk;11.6.3;GALAXY_S8;android-android;7.0;JSbridge4.4.0;jointBridge;1.1.0;"
# PCS_UA = "netdisk;P2SP;3.0.0.3;netdisk;11.5.3;PC;PC-Windows;android-android;11.0;JSbridge4.4.0"
# PCS_UA = "netdisk;P2SP;3.0.0.8;netdisk;11.12.3;GM1910;android-android;11.0;JSbridge4.4.0;jointBridge;1.1.0;"
PCS_UA = "softxm;netdisk"
PAN_UA = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/77.0.3865.75 Safari/537.36"

PCS_HEADERS = {"User-Agent": PCS_UA}
PAN_HEADERS = {"User-Agent": PAN_UA}

PCS_APP_ID = "778750"
PAN_APP_ID = "250528"

M3u8Type = Literal["M3U8_AUTO_720", "M3U8_AUTO_480"]


def _from_to(f: str, t: str) -> Dict[str, str]:
    return {"from": f, "to": t}


class Method(Enum):
    Head = "HEAD"
    Get = "GET"
    Post = "POST"


class PcsNode(Enum):
    """Pan Nodes which use pcs.baidu.com"""

    Quota = "rest/2.0/pcs/quota"
    File = "rest/2.0/pcs/file"

    def url(self) -> str:
        return f"{PCS_BAIDU_COM}/{self.value}"


class PanNode(Enum):
    """Pan Nodes which use pan.baidu.com"""

    TransferShared = "share/transfer"
    Share = "share/set"
    SharedPathList = "share/list"
    SharedRecord = "share/record"
    SharedCancel = "share/cancel"
    SharedPassword = "share/surlinfoinrecord"
    Getcaptcha = "api/getcaptcha"
    Cloud = "rest/2.0/services/cloud_dl"
    UserProducts = "rest/2.0/membership/user"

    def url(self) -> str:
        return f"{PAN_BAIDU_COM}/{self.value}"


class BaiduPCS:
    """`BaiduPCS` provides pcs's apis which return raw json"""

    def __init__(
        self,
        bduss: Optional[str] = None,
        stoken: Optional[str] = None,
        ptoken: Optional[str] = None,
        cookies: Dict[str, Optional[str]] = {},
        user_id: Optional[int] = None,
    ):
        if not bduss and cookies and cookies.get("BDUSS", ""):
            bduss = cookies["BDUSS"]
        if not stoken and cookies and cookies.get("STOKEN", ""):
            stoken = cookies["STOKEN"]
        if not ptoken and cookies and cookies.get("PTOKEN", ""):
            ptoken = cookies["PTOKEN"]

        assert bduss, "`bduss` must be set. Or `BDUSS` is in `cookies`."

        if not cookies:
            cookies = {"BDUSS": bduss, "STOKEN": stoken, "PTOKEN": ptoken}

        self._bduss = bduss
        self._stoken = stoken
        self._ptoken = ptoken
        self._bdstoken = ""
        self._logid = None
        self._baiduid = cookies.get("BAIDUID")
        if self._baiduid:
            self._logid = standard_b64encode(self._baiduid.encode("ascii")).decode(
                "utf-8"
            )

        self._cookies = cookies
        self._session = requests.Session()
        self._session.cookies.update(cookies)

        user_info = None
        if not user_id:
            user_info = self.user_info()
            user_id = user_info.get("user", {}).get("id")
            if not user_id:
                user_info = None
        self._user_id = user_id
        self._user_info = user_info

    @property
    def cookies(self) -> Dict[str, Optional[str]]:
        return self._session.cookies.get_dict()

    @staticmethod
    def _app_id(url: str):
        """Select app_id based on `url`"""

        if PCS_BAIDU_COM in url:
            return PCS_APP_ID
        else:
            return PAN_APP_ID

    @property
    def bdstoken(self) -> str:
        assert self._stoken or self._cookies.get("STOKEN")

        if self._bdstoken:
            return self._bdstoken

        url = "http://pan.baidu.com/disk/home"
        resp = self._request(Method.Get, url, params=None)
        cn = resp.text
        mod = re.search(r'bdstoken[\'":\s]+([0-9a-f]{32})', cn)
        if mod:
            s = mod.group(1)
            self._bdstoken = str(s)
            return s
        return ""

    @staticmethod
    def _headers(url: str):
        """Select headers based on `url`"""

        if PCS_BAIDU_COM in url:
            return dict(PCS_HEADERS)
        else:
            return dict(PAN_HEADERS)

    def _cookies_update(self, cookies: Dict[str, str]):
        self._session.cookies.update(cookies)

    def _request(
        self,
        method: Method,
        url: str,
        params: Optional[Dict[str, str]] = {},
        headers: Optional[Dict[str, str]] = None,
        data: Union[str, bytes, Dict[str, str], Any] = None,
        files: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> requests.Response:
        if params and isinstance(params, dict):
            app_id = self._app_id(url)
            params["app_id"] = app_id

        if not headers:
            headers = self._headers(url)

        if isinstance(data, (MultipartEncoder, MultipartEncoderMonitor)):
            assert headers
            headers["Content-Type"] = data.content_type

        try:
            resp = self._session.request(
                method.value,
                url,
                params=params,
                headers=headers,
                data=data,
                files=files,
                **kwargs,
            )
            return resp
        except Exception as err:
            raise BaiduPCSError("BaiduPCS._request", cause=err)

    def _request_get(
        self,
        url: str,
        params: Optional[Dict[str, str]] = {},
        headers: Optional[Dict[str, str]] = None,
        **kwargs,
    ) -> requests.Response:
        return self._request(Method.Get, url, params=params, headers=headers)

    @assert_ok
    def quota(self):
        """Quota space information"""

        url = PcsNode.Quota.url()
        params = {"method": "info"}
        resp = self._request(Method.Get, url, params=params)
        return resp.json()

    def meta(self, *remotepaths: str):
        assert all(
            [p.startswith("/") for p in remotepaths]
        ), "`remotepaths` must be absolute paths"

        param = [{"path": p} for p in remotepaths]
        return self.file_operate("meta", param)

    def exists(self, remotepath: str) -> bool:
        r = self.meta(remotepath)
        if r.get("error_code"):
            return False
        else:
            return True

    def is_file(self, remotepath: str) -> bool:
        r = self.meta(remotepath)
        if r.get("error_code"):
            return False
        if r["list"][0]["isdir"] == 0:
            return True
        else:
            return False

    def is_dir(self, remotepath: str) -> bool:
        r = self.meta(remotepath)
        if r.get("error_code"):
            return False
        if r["list"][0]["isdir"] == 1:
            return True
        else:
            return False

    @assert_ok
    def list(
        self,
        remotepath: str,
        desc: bool = False,
        name: bool = False,
        time: bool = False,
        size: bool = False,
    ):
        url = PcsNode.File.url()
        orderby = None
        if name:
            orderby = "name"
        elif time:
            orderby = "time"  # 服务器最后修改时间
        elif size:
            orderby = "size"
        else:
            orderby = "name"

        params = {
            "method": "list",
            "by": orderby,
            "limit": "0-2147483647",
            "order": ["asc", "desc"][desc],
            "path": str(remotepath),
        }
        resp = self._request(Method.Get, url, params=params)
        return resp.json()

    @assert_ok
    def upload_file(
        self,
        io: IO,
        remotepath: str,
        ondup="overwrite",
        callback: Callable[[MultipartEncoderMonitor], None] = None,
    ):
        """Upload the content of io to remotepath

        WARNING: This api can not set local_ctime and local_mtime
        """

        assert remotepath.startswith("/"), "`remotepath` must be an absolute path"
        remotePath = Path(remotepath)

        url = PcsNode.File.url()
        params = {
            "method": "upload",
            "ondup": ondup,
            "dir": remotePath.parent.as_posix(),
            "filename": remotePath.name,
            "BDUSS": self._bduss,
        }

        m = MultipartEncoder(fields={"file": ("file", io, "")})
        monitor = MultipartEncoderMonitor(m, callback=callback)

        resp = self._request(Method.Post, url, params=params, data=monitor)
        return resp.json()

    @assert_ok
    def rapid_upload_file(
        self,
        slice_md5: str,
        content_md5: str,
        content_crc32: int,  # not needed
        io_len: int,
        remotepath: str,
        local_ctime: Optional[int] = None,
        local_mtime: Optional[int] = None,
        ondup="overwrite",
    ):
        """Rapid Upload File

        slice_md5 (32 bytes): the md5 of pre 256KB of content.
        content_md5 (32 bytes): the md5 of total content.
        content_crc32 (int): the crc32 of total content (Not Needed),
            if content_crc32 is 0, the params of the api will be ignored.
        io_len (int): the length of total content.
        remotepath (str): the absolute remote path to save the content.
        """

        assert remotepath.startswith("/"), "`remotepath` must be an absolute path"

        url = PcsNode.File.url()
        params = {
            "method": "rapidupload",
            "BDUSS": self._bduss,
        }

        ntp = now_timestamp()

        data = {
            "path": remotepath,
            "content-length": io_len,
            "content-md5": content_md5,
            "slice-md5": slice_md5,
            "content-crc32": content_crc32,
            "local_ctime": str(local_ctime or ntp),
            "local_mtime": str(local_mtime or ntp),
            "ondup": ondup,
        }

        # Not needed
        if content_crc32 == 0:
            del data["content-crc32"]

        resp = self._request(Method.Post, url, params=params, data=data)
        return resp.json()

    @assert_ok
    def upload_slice(
        self, io: IO, callback: Callable[[MultipartEncoderMonitor], None] = None
    ):
        url = PcsNode.File.url()
        params = {
            "method": "upload",
            "type": "tmpfile",
            "BDUSS": self._bduss,
        }

        m = MultipartEncoder(fields={"file": ("file", io, "")})
        monitor = MultipartEncoderMonitor(m, callback=callback)

        resp = self._request(
            Method.Post,
            url,
            params=params,
            data=monitor,
            timeout=(3, 9),  # (connect timeout, read timeout)
        )
        return resp.json()

    @assert_ok
    def combine_slices(
        self,
        slice_md5s: List[str],
        remotepath: str,
        local_ctime: Optional[int] = None,
        local_mtime: Optional[int] = None,
        ondup="overwrite",
    ):
        url = PcsNode.File.url()
        params = {
            "method": "createsuperfile",
            "path": remotepath,
            "ondup": ondup,
            "BDUSS": self._bduss,
        }

        ntp = now_timestamp()

        data = {
            "param": dump_json({"block_list": slice_md5s}),
            "local_ctime": str(local_ctime or ntp),
            "local_mtime": str(local_mtime or ntp),
        }
        resp = self._request(Method.Post, url, params=params, data=data)
        return resp.json()

    @assert_ok
    def search(self, keyword: str, remotepath: str, recursive: bool = False):
        url = PcsNode.File.url()
        params = {
            "method": "search",
            "path": remotepath,
            "wd": keyword,
            "re": "1" if recursive else "0",
        }
        resp = self._request(Method.Get, url, params=params)
        return resp.json()

    @assert_ok
    def makedir(self, directory: str):
        url = PcsNode.File.url()
        params = {
            "method": "mkdir",
            "path": directory,
        }
        resp = self._request(Method.Get, url, params=params)
        return resp.json()

    def file_operate(self, operate: str, param: List[Dict[str, str]]):
        url = PcsNode.File.url()
        params = {"method": operate}
        data = {"param": dump_json({"list": param})}
        resp = self._request(Method.Post, url, params=params, data=data)
        return resp.json()

    @assert_ok
    def move(self, *remotepaths: str):
        """
        Move sources to destination

        sources, dest = remotepaths[:-1], remotepaths[-1]

        `dest` must be a directory
        """

        assert len(remotepaths) > 1 and all(
            [p.startswith("/") for p in remotepaths]
        ), "`sources`, `dest` must be absolute paths"

        sources, dest = remotepaths[:-1], remotepaths[-1]

        if self.is_file(dest):
            raise BaiduPCSError("The remote `dest` is a file. It must be a directory.")

        if not self.is_dir(dest):
            self.makedir(dest)

        _sources = (Path(s) for s in sources)
        _dest = Path(dest)

        param = [_from_to(s.as_posix(), (_dest / s.name).as_posix()) for s in _sources]
        return self.file_operate("move", param)

    @assert_ok
    def rename(self, source: str, dest: str):
        """Rename `source` to `dest`"""

        assert all(
            [p.startswith("/") for p in [source, dest]]
        ), "`source`, `dest` must be absolute paths"

        param = [_from_to(source, dest)]
        return self.file_operate("move", param)

    @assert_ok
    def copy(self, *remotepaths: str):
        """
        Copy sources to destination

        sources, dest = remotepaths[:-1], remotepaths[-1]

        `dest` must be a directory
        """

        assert len(remotepaths) > 1 and all(
            [p.startswith("/") for p in remotepaths]
        ), "`sources`, `dest` must be absolute paths"

        sources, dest = remotepaths[:-1], remotepaths[-1]

        if self.is_file(dest):
            raise BaiduPCSError("The remote `dest` is a file. It must be a directory.")

        if not self.is_dir(dest):
            self.makedir(dest)

        _sources = (Path(s) for s in sources)
        _dest = Path(dest)

        param = [_from_to(s.as_posix(), (_dest / s.name).as_posix()) for s in _sources]
        return self.file_operate("copy", param)

    @assert_ok
    def remove(self, *remotepaths: str):
        assert all(
            [p.startswith("/") for p in remotepaths]
        ), "`sources`, `dest` must be absolute paths"

        param = [{"path": p} for p in remotepaths]
        return self.file_operate("delete", param)

    @assert_ok
    def cloud_operate(
        self, params: Dict[str, str], data: Optional[Dict[str, str]] = None
    ):
        url = PanNode.Cloud.url()
        if data:
            resp = self._request(Method.Post, url, params=params, data=data)
        else:
            resp = self._request(Method.Get, url, params=params)
        return resp.json()

    def magnet_info(self, magnet: str):
        params = {
            "method": "query_magnetinfo",
            "source_url": magnet,
            # 'save_path': '/',
            "type": "4",
        }
        return self.cloud_operate(params)

    def torrent_info(self, remote_torrent: str):
        params = {
            "method": "query_sinfo",
            "source_path": remote_torrent,
            # 'save_path': '/',
            "type": "2",
            # 't': int(time.time() * 1000),
        }
        return self.cloud_operate(params)

    def add_task(self, task_url: str, remotedir: str):
        """Add cloud task for http/s and ed2k url

        Warning: `STOKEN` must be in `cookies`
        """

        assert self._stoken, "`STOKEN` is not in `cookies`"

        params = {
            "channel": "chunlei",
            "web": "1",
            "app_id": "250528",
            "bdstoken": self.bdstoken,  # Must be set
            # "logid": self._logid,
            "clienttype": "0",
        }
        data = {
            "method": "add_task",
            "app_id": "250528",
            "save_path": remotedir,
            "source_url": task_url,
            "timeout": "2147483647",
        }
        return self.cloud_operate(params, data=data)

    def add_magnet_task(self, task_url: str, remotedir: str, selected_idx: List[int]):
        """Add cloud task for magnet

        Args:
            selected_idx (List[int]): indexes of `BaiduPCS.magnet_info` list,
                starting from 1

        Warning: `STOKEN` must be in `cookies`
        """

        assert self._stoken, "`STOKEN` is not in `cookies`"

        params = {
            "channel": "chunlei",
            "web": "1",
            "app_id": "250528",
            "bdstoken": self.bdstoken,  # Must be set
            # "logid": self._logid,
            "clienttype": "0",
        }
        data = {
            "method": "add_task",
            "app_id": "250528",
            "save_path": remotedir,
            "source_url": task_url,
            "timeout": "2147483647",
            "type": "4",
            "t": str(int(time.time() * 1000)),
            "file_sha1": "",
            "selected_idx": ",".join([str(i) for i in selected_idx or []]),
            "task_from": "1",
        }
        return self.cloud_operate(params, data=data)

    def tasks(self, *task_ids: str):
        params = {
            "method": "query_task",
            "task_ids": ",".join(task_ids),
        }
        return self.cloud_operate(params)

    def list_tasks(self):
        params = {
            "method": "list_task",
            "need_task_info": "1",
            "status": "255",
            "start": "0",
            "limit": "1000",
        }
        return self.cloud_operate(params)

    def clear_tasks(self):
        params = {
            "method": "clear_task",
        }
        return self.cloud_operate(params)

    def cancel_task(self, task_id: str):
        params = {
            "method": "cancel_task",
            "task_id": task_id,
        }
        return self.cloud_operate(params)

    @assert_ok
    def share(self, *remotepaths: str, password: str, period: int = 0):
        """Share `remotepaths` to public

        period (int): The days for expiring. `0` means no expiring
        """

        assert self._stoken, "`STOKEN` is not in `cookies`"
        assert len(password) == 4, "`password` MUST be set"

        meta = self.meta(*remotepaths)
        fs_ids = [i["fs_id"] for i in meta["list"]]

        url = PanNode.Share.url()
        params = {
            "channel": "chunlei",
            "clienttype": "0",
            "web": "1",
            "bdstoken": self.bdstoken,
        }
        data = {
            "fid_list": dump_json(fs_ids),
            "schannel": "0",
            "channel_list": "[]",
            "period": str(int(period)),
        }
        if password:
            data["pwd"] = password
            data["schannel"] = "4"

        resp = self._request(Method.Post, url, params=params, data=data)
        return resp.json()

    @assert_ok
    def list_shared(self, page: int = 1):
        """
        list.0.channel:
            - 0, no password
            - 4, with password
        """

        url = PanNode.SharedRecord.url()
        params = {
            "page": str(page),
            "desc": "1",
            "order": "time",
        }
        resp = self._request(Method.Get, url, params=params)
        return resp.json()

    @assert_ok
    def shared_password(self, share_id: int):
        """
        Only return password
        """

        url = PanNode.SharedPassword.url()
        params = {
            "shareid": str(share_id),
            "sign": calu_md5(f"{share_id}_sharesurlinfo!@#"),
        }
        resp = self._request(Method.Get, url, params=params)
        return resp.json()

    @assert_ok
    def cancel_shared(self, *share_ids: int):
        url = PanNode.SharedCancel.url()
        data = {
            "shareid_list": dump_json(share_ids),
        }
        hdrs = dict(PCS_HEADERS)
        hdrs["Content-Type"] = "application/x-www-form-urlencoded"
        resp = self._request(Method.Post, url, headers=hdrs, params=None, data=data)
        return resp.json()

    def shared_init_url(self, shared_url: str) -> str:
        u = urlparse(shared_url)
        surl = u.path.split("/s/1")[-1]
        return f"https://pan.baidu.com/share/init?surl={surl}"

    @assert_ok
    def access_shared(self, shared_url: str, password: str, vcode_str: str, vcode: str):
        """Pass password to the session

        WARNING: this method is not threadsafe.
        """

        url = "https://pan.baidu.com/share/verify"
        init_url = self.shared_init_url(shared_url)
        params = {
            "surl": init_url.split("surl=")[-1],
            "t": str(now_timestamp() * 1000),
            "channel": "chunlei",
            "web": "1",
            "bdstoken": "null",
            "clienttype": "0",
        }
        data = {
            "pwd": password,
            "vcode": vcode,
            "vcode_str": vcode_str,
        }
        hdrs = dict(PAN_HEADERS)
        hdrs["Referer"] = init_url
        resp = self._request(Method.Post, url, headers=hdrs, params=params, data=data)

        # These cookies must be included through all sub-processes
        self._cookies_update(resp.cookies.get_dict())

        return resp.json()

    @assert_ok
    def getcaptcha(self, shared_url: str) -> str:
        url = PanNode.Getcaptcha.url()
        params = {
            "prod": "shareverify",
            "channel": "chunlei",
            "web": "1",
            "bdstoken": "null",
            "clienttype": "0",
        }

        hdrs = dict(PAN_HEADERS)
        hdrs["Referer"] = self.shared_init_url(shared_url)
        hdrs["X-Requested-With"] = "XMLHttpRequest"
        hdrs["Content-Type"] = "application/x-www-form-urlencoded; charset=UTF-8"
        resp = self._request(Method.Get, url, headers=hdrs, params=params)
        return resp.json()

    def get_vcode_img(self, vcode_img_url: str, shared_url: str) -> bytes:
        hdrs = dict(PAN_HEADERS)
        hdrs["Referer"] = self.shared_init_url(shared_url)
        resp = self._request_get(vcode_img_url, headers=hdrs)
        return resp.content

    @assert_ok
    def shared_paths(self, shared_url: str):
        """Get shared paths

        Call `BaiduPCS.access_share` before calling the function

        WARNING: this method is not threadsafe.
        """

        assert self._stoken, "`STOKEN` is not in `cookies`"

        resp = self._request(Method.Get, shared_url, params=None)
        html = resp.text

        # These cookies must be included through all sub-processes
        self._cookies_update(resp.cookies.get_dict())

        m = re.search(r"(?:yunData.setData|locals.mset)\((.+?)\);", html)
        assert m, "`BaiduPCS.shared_paths`: Don't get shared info"

        shared_data = m.group(1)
        return json.loads(shared_data)

    @assert_ok
    def list_shared_paths(
        self, sharedpath: str, uk: int, share_id: int, page: int = 1, size: int = 100
    ):
        assert self._stoken, "`STOKEN` is not in `cookies`"

        url = PanNode.SharedPathList.url()
        params = {
            "channel": "chunlei",
            "clienttype": "0",
            "web": "1",
            "page": str(page),  # from 1
            "num": str(size),  # max is 100
            "dir": sharedpath,
            "t": str(random.random()),
            "uk": str(uk),
            "shareid": str(share_id),
            "desc": "1",  # reversely
            "order": "other",  # sort by name, or size, time
            "bdstoken": "null",
            "showempty": "0",
        }
        resp = self._request(Method.Get, url, params=params)
        return resp.json()

    @assert_ok
    def transfer_shared_paths(
        self,
        remotedir: str,
        fs_ids: List[int],
        uk: int,
        share_id: int,
        bdstoken: str,
        shared_url: str,
    ):
        """`remotedir` must exist"""

        url = PanNode.TransferShared.url()
        params = {
            "shareid": str(share_id),
            "from": str(uk),
            "bdstoken": bdstoken,
            "channel": "chunlei",
            "clienttype": "0",
            "web": "1",
        }
        data = {
            "fsidlist": dump_json(fs_ids),
            "path": remotedir,
        }
        hdrs = dict(PAN_HEADERS)
        hdrs["X-Requested-With"] = "XMLHttpRequest"
        hdrs["Content-Type"] = "application/x-www-form-urlencoded; charset=UTF-8"
        hdrs["Origin"] = "https://pan.baidu.com"
        hdrs["Referer"] = shared_url  # WARNING: Referer must be set to shared_url

        resp = self._request(Method.Post, url, headers=hdrs, params=params, data=data)
        info = resp.json()
        if info.get("info") and info["info"][0]["errno"]:
            info["errno"] = info["info"][0]["errno"]
        return info

    @assert_ok
    def user_info(self):
        bduss = self._bduss
        timestamp = str(now_timestamp())
        model = get_phone_model(bduss)
        phoneIMEIStr = sum_IMEI(bduss)

        data = {
            "bdusstoken": bduss + "|null",
            "channel_id": "",
            "channel_uid": "",
            "stErrorNums": "0",
            "subapp_type": "mini",
            "timestamp": timestamp + "922",
        }
        data["_client_type"] = "2"
        data["_client_version"] = "7.0.0.0"
        data["_phone_imei"] = phoneIMEIStr
        data["from"] = "mini_ad_wandoujia"
        data["model"] = model
        data["cuid"] = (
            calu_md5(
                bduss
                + "_"
                + data["_client_version"]
                + "_"
                + data["_phone_imei"]
                + "_"
                + data["from"]
            ).upper()
            + "|"
            + phoneIMEIStr[::-1]
        )
        data["sign"] = calu_md5(
            "".join([k + "=" + data[k] for k in sorted(data.keys())]) + "tiebaclient!!!"
        ).upper()

        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Cookie": "ka=open",
            "net": "1",
            "User-Agent": "bdtb for Android 6.9.2.1",
            "client_logid": timestamp + "416",
            "Connection": "Keep-Alive",
        }

        resp = requests.post(
            "http://tieba.baidu.com/c/s/login", headers=headers, data=data
        )
        return resp.json()

    @assert_ok
    def tieba_user_info(self, user_id: int):
        params = f"has_plist=0&need_post_count=1&rn=1&uid={user_id}"
        params += "&sign=" + calu_md5(params.replace("&", "") + "tiebaclient!!!")
        url = "http://c.tieba.baidu.com/c/u/user/profile?" + params
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Cookie": "ka=open",
            "net": "1",
            "User-Agent": "bdtb for Android 6.9.2.1",
            "client_logid": str(now_timestamp() * 1000),
            "Connection": "Keep-Alive",
        }
        resp = requests.get(url, headers=headers, params=None)
        return resp.json()

    @assert_ok
    def user_products(self):
        url = PanNode.UserProducts.url()
        params = {
            "method": "query",
        }
        resp = self._request(Method.Get, url, params=params)
        return resp.json()

    @timeout_cache(1 * 60 * 60)  # 1 hour timeout
    def download_link(self, remotepath: str, pcs: bool = False) -> Optional[str]:
        if pcs:
            return (
                "http://c.pcs.baidu.com/rest/2.0/pcs/file"
                f"?method=download&app_id={PCS_APP_ID}&path={quote_plus(remotepath)}"
                "&ver=2.0&clienttype=1"
            )

        bduss = self._bduss
        uid = str(self._user_id or "")
        devuid = calu_md5(bduss).upper() + "|0"
        enc = calu_sha1(bduss)

        while True:
            timestamp = str(now_timestamp())

            rand = calu_sha1(
                enc + uid + "ebrcUYiuxaZv2XGu7KIYKxUrqfnOfpDF" + timestamp + devuid
            )

            url = PcsNode.File.url()
            params = {
                "apn_id": "1_0",
                "app_id": PAN_APP_ID,
                "channel": "0",
                "check_blue": "1",
                "clienttype": "17",
                "es": "1",
                "esl": "1",
                "freeisp": "0",
                "method": "locatedownload",
                "path": quote_plus(remotepath),
                "queryfree": "0",
                "use": "0",
                "ver": "4.0",
                "time": timestamp,
                "rand": rand,
                "devuid": devuid,
                "cuid": devuid,
            }

            params_str = "&".join([f"{k}={v}" for k, v in params.items()])

            headers = dict(PCS_HEADERS)
            headers["Cookie"] = "; ".join([f"{k}={v}" for k, v in self.cookies.items()])
            req = urllib.request.Request(
                url + "?" + params_str, headers=headers, method="GET"
            )  # type: ignore
            try:
                resp = urllib.request.urlopen(req)  # type: ignore

                # Error: "user is not authorized"
                # This error occurs when the method is called by too many times
                if resp.status != 200:
                    time.sleep(2)
                    continue

                info = json.loads(resp.read())

                # This error is gotten when remote path is blocked
                if info.get("host") == "issuecdn.baidupcs.com":
                    return None

                if not info.get("urls"):
                    return None
                else:
                    # return info["urls"][0]["url"].replace("&htype=", "")
                    return info["urls"][0]["url"]
            except HTTPError:
                # 403 code could occor at unavailable downloading url, #97
                return None

    def file_stream(
        self,
        remotepath: str,
        max_chunk_size: int = MAX_CHUNK_SIZE,
        callback: Callable[..., None] = None,
        encrypt_password: bytes = b"",
        pcs: bool = False,
    ) -> Optional[RangeRequestIO]:
        url = self.download_link(remotepath, pcs=pcs)
        if not url:
            return None

        headers = {
            "Cookie": f"BDUSS={self._bduss};",
            "User-Agent": PCS_UA,
            "Connection": "Keep-Alive",
        }
        return RangeRequestIO(
            Method.Get.value,
            url,
            headers=headers,
            max_chunk_size=max_chunk_size,
            callback=callback,
            encrypt_password=encrypt_password,
        )

    # Playing the m3u8 file is needed add `--stream-lavf-o-append="protocol_whitelist=file,http,https,tcp,tls,crypto,hls,applehttp"` for mpv
    # https://github.com/mpv-player/mpv/issues/6928#issuecomment-532198445
    @assert_ok
    def m3u8_stream(self, remotepath: str, type: M3u8Type = "M3U8_AUTO_720"):
        """Get content of the m3u8 stream file"""

        url = PcsNode.File.url()
        params = {
            "method": "streaming",
            "path": remotepath,
            "type": type,
        }

        resp = self._request(Method.Get, url, params=params)
        cn = resp.text
        if cn.startswith("{"):  # error
            return resp.json()
        else:
            return {"m3u8_content": cn}



from functools import wraps
from typing import Any, Optional

ERRORS = {
    0: 0,
    -1: "由于您分享了违反相关法律法规的文件，分享功能已被禁用，之前分享出去的文件不受影响。",
    -2: "用户不存在,请刷新页面后重试",
    -3: "文件不存在,请刷新页面后重试",
    -4: "登录信息有误，请重新登录试试",
    -5: "host_key和user_key无效",
    -6: "请重新登录",
    -7: "该分享已删除或已取消",
    -8: "该分享已经过期",
    -9: "文件不存在 或 提取密码错误",
    -10: "分享外链已经达到最大上限100000条，不能再次分享",
    -11: "验证cookie无效",
    -12: "访问密码错误",
    -14: "对不起，短信分享每天限制20条，你今天已经分享完，请明天再来分享吧！",
    -15: "对不起，邮件分享每天限制20封，你今天已经分享完，请明天再来分享吧！",
    -16: "对不起，该文件已经限制分享！",
    -17: "文件分享超过限制",
    -19: "需要输入验证码",
    -21: "分享已取消或分享信息无效",
    -30: "文件已存在",
    -31: "文件保存失败",
    -32: "剩余空间不足，无法转存",
    -33: "一次支持操作999个，减点试试吧",
    -62: "可能需要输入验证码",
    -70: "你分享的文件中包含病毒或疑似病毒，为了你和他人的数据安全，换个文件分享吧",
    2: "参数错误",
    3: "未登录或帐号无效",
    4: "存储好像出问题了，请稍候再试",
    12: "文件已经存在",
    105: "啊哦，链接错误没找到文件，请打开正确的分享链接",
    108: "文件名有敏感词，优化一下吧",
    110: "分享次数超出限制，可以到“我的分享”中查看已分享的文件链接",
    112: "页面已过期，请刷新后重试",
    113: "签名错误",
    114: "当前任务不存在，保存失败",
    115: "该文件禁止分享",
    130: "转存文件数超限",
    132: "您的帐号可能存在安全风险，为了确保为您本人操作，请先进行安全验证。",
    31001: "数据库查询错误",
    31002: "数据库连接错误",
    31003: "数据库返回空结果",
    31021: "网络错误",
    31022: "暂时无法连接服务器",
    31023: "输入参数错误",
    31024: "app id为空",
    31025: "后端存储错误",
    31041: "用户的cookie不是合法的百度cookie",
    31042: "用户未登陆",
    31043: "用户未激活",
    31044: "用户未授权",
    31045: "用户不存在",
    31046: "用户已经存在",
    31061: "文件已经存在",
    31062: "文件名非法",
    31063: "文件父目录不存在",
    31064: "无权访问此文件",
    31065: "目录已满",
    31066: "文件不存在",
    31067: "文件处理出错",
    31068: "文件创建失败",
    31069: "文件拷贝失败",
    31070: "文件删除失败",
    31071: "不能读取文件元信息",
    31072: "文件移动失败",
    31073: "文件重命名失败",
    31081: "superfile创建失败",
    31082: "superfile 块列表为空",
    31083: "superfile 更新失败",
    31101: "tag系统内部错误",
    31102: "tag参数错误",
    31103: "tag系统错误",
    31110: "未授权设置此目录配额",
    31111: "配额管理只支持两级目录",
    31112: "超出配额",
    31113: "配额不能超出目录祖先的配额",
    31114: "配额不能比子目录配额小",
    31141: "请求缩略图服务失败",
    31201: "签名错误",
    31203: "设置acl失败",
    31204: "请求acl验证失败",
    31205: "获取acl失败",
    31079: "未找到文件MD5，请使用上传API上传整个文件。",
    31202: "文件不存在",
    31206: "acl不存在",
    31207: "bucket已存在",
    31208: "用户请求错误",
    31209: "服务器错误",
    31210: "服务器不支持",
    31211: "禁止访问",
    31212: "服务不可用",
    31213: "重试出错",
    31214: "上传文件data失败",
    31215: "上传文件meta失败",
    31216: "下载文件data失败",
    31217: "下载文件meta失败",
    31218: "容量超出限额",
    31219: "请求数超出限额",
    31220: "流量超出限额",
    31298: "服务器返回值KEY非法",
    31299: "服务器返回值KEY不存在",
    31304: "file type is not supported",
    31341: "be transcoding, please wait and retry",
    31326: "user is not authorized, hitcode:117",  # Hit when request download link
    31626: "user is not authorized, hitcode:122",  # Hit when request download link
    36001: "离线下载错误",
    36032: "hit sexy spam",
    36037: "source url not support",
    100001: "文件被封",
}

UNKNOWN_ERROR = "未知错误"


class BaiduPCSError(Exception):
    def __init__(self, message: str, error_code: Optional[int] = None, cause=None):
        self.__cause__ = cause
        self.error_code = error_code
        super().__init__(message)


def parse_errno(error_code: int, info: Any = None) -> Optional[BaiduPCSError]:
    if error_code != 0:
        mean = ERRORS.get(error_code, info or UNKNOWN_ERROR)
        msg = f"error_code: {error_code}, message: {mean}"
        return BaiduPCSError(msg, error_code=error_code)
    return None


def assert_ok(func):
    """Assert the errno of response is not 0"""

    @wraps(func)
    def check(*args, **kwargs):
        info = func(*args, **kwargs)
        error_code = info.get("errno")
        if error_code is None:
            error_code = info.get("error_code")
        if error_code is None:
            error_code = 0
        error_code = int(error_code)

        if error_code not in ERRORS:
            err = parse_errno(error_code, str(info))
        else:
            err = parse_errno(error_code)

        if err:
            raise err
        return info

    return check



from typing import Optional, List, Dict, Any

import os
from collections import OrderedDict

import sqlite3


RAPID_UPLOAD_TABLE = "rapid_upload"

CREATE_RAPID_UPLOAD_TABLE = f"""
CREATE TABLE IF NOT EXISTS {RAPID_UPLOAD_TABLE}
(
    id INTEGER PRIMARY KEY AUTOINCREMENT,

    filename TEXT NULL,

    localpath TEXT NULL,
    remotepath TEXT NULL,

    encrypt_password TEXT NULL,
    encrypt_type TEXT NULL,

    user_id INTEGER NULL,
    user_name TEXT NULL,

    slice_md5 TEXT NOT NULL,
    content_md5 TEXT NOT NULL,
    content_crc32 TEXT NOT NULL,
    content_length INTEGER NOT NULL,

    record_time TIMESTAMP DEFAULT CURRENT_TIMESTAMP,

    UNIQUE(slice_md5, content_md5, content_length, filename)
)
"""

RAPID_UPLOAD_TABLE_COLS = [
    c.strip().split(" ", 1)[0]
    for c in CREATE_RAPID_UPLOAD_TABLE.split("\n")
    if c.startswith("    ")
]

INSERT_RAPID_UPLOAD = f"""
INSERT OR IGNORE INTO {RAPID_UPLOAD_TABLE}
    (
        filename,

        localpath,
        remotepath,

        encrypt_password,
        encrypt_type,

        user_id,
        user_name,

        slice_md5,
        content_md5,
        content_crc32,
        content_length
    )
VALUES (?,?,?,?,?,?,?,?,?,?,?)
"""


class RapidUploadInfo:
    def __init__(self, db_path: str):
        self._db_path = db_path
        self._conn = sqlite3.connect(db_path)

        c = self._conn.cursor()
        c.execute(CREATE_RAPID_UPLOAD_TABLE)
        self._conn.commit()

    def insert(
        self,
        slice_md5: str,
        content_md5: str,
        content_crc32: int,  # not needed
        content_length: int,
        filename: Optional[str] = None,
        localpath: Optional[str] = None,
        remotepath: Optional[str] = None,
        encrypt_password: Optional[bytes] = None,
        encrypt_type: Optional[str] = None,
        user_id: Optional[int] = None,
        user_name: Optional[str] = None,
    ):
        """Insert a rapid upload info"""

        c = self._conn.cursor()
        c.execute(
            INSERT_RAPID_UPLOAD,
            (
                filename,
                localpath,
                remotepath,
                encrypt_password,
                encrypt_type,
                user_id,
                user_name,
                slice_md5,
                content_md5,
                content_crc32,
                content_length,
            ),
        )
        self._conn.commit()

    def list(
        self,
        ids: List[int] = [],
        by_filename: bool = False,
        by_time: bool = False,
        by_size: bool = False,
        by_localpath: bool = False,
        by_remotepath: bool = False,
        by_user_id: bool = False,
        by_user_name: bool = False,
        desc: bool = False,
        limit: int = 0,
        offset: int = -1,
    ) -> List[Dict[str, Any]]:
        """List records by condition

        Default order by record_time and desc
        """

        if ids:
            ids_str = ",".join([str(i) for i in ids])
            sql = f"SELECT * FROM {RAPID_UPLOAD_TABLE} WHERE id IN ({ids_str})"
            c = self._conn.cursor()
            c.execute(sql)
            return [OrderedDict(zip(RAPID_UPLOAD_TABLE_COLS, r)) for r in c.fetchall()]

        if by_filename:
            condition = "order by filename"
        elif by_time:
            condition = "order by record_time"
        elif by_size:
            condition = "order by content_length"
        elif by_localpath:
            condition = "order by localpath"
        elif by_remotepath:
            condition = "order by remotepath"
        elif by_user_id:
            condition = "order by user_id"
        elif by_user_name:
            condition = "order by user_name"
        else:
            condition = "order by record_time"
            desc = True

        if desc:
            condition += " desc"

        if limit > 0:
            condition += f" LIMIT {limit}"

        if offset > 0:
            condition += f" OFFSET {offset}"

        sql = f"SELECT * FROM {RAPID_UPLOAD_TABLE} {condition}"

        c = self._conn.cursor()
        c.execute(sql)
        return [OrderedDict(zip(RAPID_UPLOAD_TABLE_COLS, r)) for r in c.fetchall()]

    def search(
        self,
        keyword: str,
        in_filename: bool = False,
        in_localpath: bool = False,
        in_remotepath: bool = False,
        in_user_name: bool = False,
        in_md5: bool = False,
    ) -> List[Dict[str, Any]]:
        keyword = keyword.replace("'", r"\'")

        conditions = []
        if in_filename:
            conditions.append(f"filename LIKE '%{keyword}%'")
        if in_localpath:
            conditions.append(f"localpath LIKE '%{keyword}%'")
        if in_remotepath:
            conditions.append(f"remotepath LIKE '%{keyword}%'")
        if in_user_name:
            conditions.append(f"user_name LIKE '%{keyword}%'")
        if in_md5:
            conditions.append(f"content_md5 LIKE '%{keyword}%'")

        if not conditions:
            conditions = [
                f"filename LIKE '%{keyword}%'",
                f"localpath LIKE '%{keyword}%'",
                f"remotepath LIKE '%{keyword}%'",
                f"user_name LIKE '%{keyword}%'",
                f"content_md5 LIKE '%{keyword}%'",
            ]

        condition = "WHERE " + " OR ".join(conditions)

        if keyword:
            sql = f"SELECT * FROM {RAPID_UPLOAD_TABLE} {condition}"
        else:
            sql = f"SELECT * FROM {RAPID_UPLOAD_TABLE}"

        c = self._conn.cursor()
        c.execute(sql)
        return [OrderedDict(zip(RAPID_UPLOAD_TABLE_COLS, r)) for r in c.fetchall()]

    def delete(self, id: int):
        sql = f"DELETE FROM {RAPID_UPLOAD_TABLE} WHERE id = ?"

        c = self._conn.cursor()
        c.execute(sql, (id,))
        self._conn.commit()


def save_rapid_upload_info(
    rapiduploadinfo_file: str,
    slice_md5: str,
    content_md5: str,
    content_crc32: int,  # not needed
    content_length: int,
    localpath: Optional[str] = None,
    remotepath: Optional[str] = None,
    encrypt_password: bytes = b"",
    encrypt_type: str = "",
    user_id: Optional[int] = None,
    user_name: Optional[str] = None,
):
    rapiduploadinfo = RapidUploadInfo(rapiduploadinfo_file)
    rapiduploadinfo.insert(
        slice_md5.lower(),
        content_md5.lower(),
        content_crc32,
        content_length,
        filename=os.path.basename(remotepath or ""),
        localpath=localpath,
        remotepath=remotepath,
        encrypt_password=encrypt_password,
        encrypt_type=encrypt_type,
        user_id=user_id,
        user_name=user_name,
    )



import os
from enum import Enum
from typing import List

from rich import print

from ..common.file_type import ARCHIVE_EXTS, DOC_EXTS, IMAGE_EXTS, MEDIA_EXTS
from ..common.url import is_magnet
from ..pcs import BaiduPCSApi
from .display import display_tasks
from .log import get_logger

logger = get_logger(__name__)


class FileType(Enum):
    All = "All"
    Media = "Media"
    Image = "Image"
    Doc = "Doc"
    Archive = "Archive"

    def sift(self, ext: str) -> bool:
        ext = ext.lower()
        if self == FileType.Media:
            if ext in MEDIA_EXTS:
                return True
        elif self == FileType.Image:
            if ext in IMAGE_EXTS:
                return True
        elif self == FileType.Doc:
            if ext in DOC_EXTS:
                return True
        elif self == FileType.Archive:
            if ext in ARCHIVE_EXTS:
                return True
        elif self == FileType.All:
            return True
        else:
            raise RuntimeError("Here is unreachable")

        return False

    @staticmethod
    def from_(addr: str) -> "FileType":
        if addr == "a":
            return FileType.All
        if addr == "m":
            return FileType.Media
        if addr == "i":
            return FileType.Image
        if addr == "d":
            return FileType.Doc
        if addr == "c":
            return FileType.Archive

        logger.warning("Unknown FileType addr: %s", addr)

        raise ValueError(f"Unknown FileType addr: {addr}")


def add_task(
    api: BaiduPCSApi,
    task_url: str,
    remotedir: str,
    file_types: List[FileType] = [FileType.Media],
):
    if is_magnet(task_url):
        logger.warning("Add cloud task: magnet: %s", task_url)

        pmfs = api.magnet_info(task_url)
        selected_idx = []
        for idx, pmf in enumerate(pmfs, 1):
            ext = os.path.splitext(pmf.path)[-1]
            for file_type in file_types:
                if file_type.sift(ext):
                    selected_idx.append(idx)
                    break

        if not selected_idx:
            logger.warning("`add_task`: No selected idx for %s", task_url)
            return

        task_id = api.add_magnet_task(task_url, remotedir, selected_idx)
        tasks(api, task_id)
    else:
        logger.warning("Add cloud task: %s", task_url)
        task_id = api.add_task(task_url, remotedir)


def tasks(api: BaiduPCSApi, *task_ids: str):
    if not task_ids:
        return

    cloud_tasks = api.tasks(*task_ids)
    display_tasks(*cloud_tasks)


def list_tasks(api: BaiduPCSApi):
    cloud_tasks = api.list_tasks()
    tasks(api, *[t.task_id for t in cloud_tasks])


def clear_tasks(api: BaiduPCSApi):
    n = api.clear_tasks()
    print("clear tasks:", n)


def cancel_task(api: BaiduPCSApi, task_id: str):
    api.cancel_task(task_id)


def purge_all_tasks(api: BaiduPCSApi):
    clear_tasks(api)

    cloud_tasks = api.list_tasks()
    for task in cloud_tasks:
        cancel_task(api, task.task_id)



from concurrent.futures import ThreadPoolExecutor, as_completed
from threading import Semaphore
from typing import List, Optional, Tuple

from rich import print

from ..common.localstorage import save_rapid_upload_info
from ..pcs import BaiduPCSApi
from ..pcs.inner import PcsFile, PcsRapidUploadInfo
from .cat.common.concurrent import sure_release
from .display import display_files
from .log import get_logger
from .sifter import Sifter, sift

logger = get_logger(__name__)

DEFAULT_MAX_WORKERS = 10


def _get_download_link_and_rapid_upload_info(
    api: BaiduPCSApi,
    pcs_file: PcsFile,
    show_dl_link: bool = False,
    show_hash_link: bool = False,
    check_md5: bool = True,
) -> Tuple[Optional[str], Optional[PcsRapidUploadInfo]]:
    dl_link = None
    if show_dl_link:
        dl_link = api.download_link(pcs_file.path)

    rpinfo = None
    if show_hash_link:
        rpinfo = api.rapid_upload_info(pcs_file.path, check=check_md5)

    return dl_link, rpinfo


def list_file(
    api: BaiduPCSApi,
    remotepath: str,
    desc: bool = False,
    name: bool = False,
    time: bool = False,
    size: bool = False,
    recursive: bool = False,
    sifters: List[Sifter] = [],
    highlight: bool = False,
    rapiduploadinfo_file: Optional[str] = None,
    user_id: Optional[int] = None,
    user_name: Optional[str] = None,
    show_size: bool = False,
    show_date: bool = False,
    show_md5: bool = False,
    show_absolute_path: bool = False,
    show_dl_link: bool = False,
    show_hash_link: bool = False,
    hash_link_protocol: str = PcsRapidUploadInfo.default_hash_link_protocol(),
    check_md5: bool = True,
    csv: bool = False,
    only_dl_link: bool = False,
    only_hash_link: bool = False,
):
    is_dir = api.is_dir(remotepath)
    if is_dir:
        pcs_files = api.list(remotepath, desc=desc, name=name, time=time, size=size)
    else:
        pcs_files = api.meta(remotepath)

    pcs_files = sift(pcs_files, sifters, recursive=recursive)
    if not pcs_files:
        return

    if show_dl_link or show_hash_link:
        # Concurrently request rapiduploadinfo
        max_workers = DEFAULT_MAX_WORKERS
        semaphore = Semaphore(max_workers)
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futs = {}
            for i in range(len(pcs_files)):
                if pcs_files[i].is_dir:
                    continue

                semaphore.acquire()
                fut = executor.submit(
                    sure_release,
                    semaphore,
                    _get_download_link_and_rapid_upload_info,
                    api,
                    pcs_files[i],
                    show_dl_link=show_dl_link,
                    show_hash_link=show_hash_link,
                    check_md5=check_md5,
                )
                futs[fut] = i

            for fut in as_completed(futs):
                i = futs[fut]
                e = fut.exception()
                if e is None:
                    dl_link, rpinfo = fut.result()
                    if rapiduploadinfo_file and rpinfo:
                        save_rapid_upload_info(
                            rapiduploadinfo_file,
                            rpinfo.slice_md5,
                            rpinfo.content_md5,
                            rpinfo.content_crc32,
                            rpinfo.content_length,
                            remotepath=pcs_files[i].path,
                            user_id=user_id,
                            user_name=user_name,
                        )
                    pcs_files[i] = pcs_files[i]._replace(
                        dl_link=dl_link, rapid_upload_info=rpinfo
                    )
                    if only_dl_link and dl_link:
                        print(dl_link)
                    if only_hash_link and rpinfo:
                        hash_link = getattr(rpinfo, hash_link_protocol)()
                        print(hash_link)
                else:
                    logger.error(
                        "`list_file`: `_get_download_link_and_rapid_upload_info` error: %s",
                        e,
                    )

    if not only_dl_link and not only_hash_link:
        display_files(
            pcs_files,
            remotepath,
            sifters=sifters,
            highlight=highlight,
            show_size=show_size,
            show_date=show_date,
            show_md5=show_md5,
            show_absolute_path=show_absolute_path,
            show_dl_link=show_dl_link,
            show_hash_link=show_hash_link,
            hash_link_protocol=hash_link_protocol,
            csv=csv,
        )

    if is_dir and recursive:
        for pcs_file in pcs_files:
            if pcs_file.is_dir:
                list_file(
                    api,
                    pcs_file.path,
                    desc=desc,
                    name=name,
                    time=time,
                    size=size,
                    recursive=recursive,
                    sifters=sifters,
                    highlight=highlight,
                    rapiduploadinfo_file=rapiduploadinfo_file,
                    user_id=user_id,
                    user_name=user_name,
                    show_size=show_size,
                    show_date=show_date,
                    show_md5=show_md5,
                    show_absolute_path=show_absolute_path,
                    show_dl_link=show_dl_link,
                    show_hash_link=show_hash_link,
                    hash_link_protocol=hash_link_protocol,
                    check_md5=check_md5,
                    csv=csv,
                    only_dl_link=only_dl_link,
                    only_hash_link=only_hash_link,
                )


def list_files(
    api: BaiduPCSApi,
    *remotepaths: str,
    desc: bool = False,
    name: bool = False,
    time: bool = False,
    size: bool = False,
    recursive: bool = False,
    sifters: List[Sifter] = [],
    highlight: bool = False,
    rapiduploadinfo_file: Optional[str] = None,
    user_id: Optional[int] = None,
    user_name: Optional[str] = None,
    show_size: bool = False,
    show_date: bool = False,
    show_md5: bool = False,
    show_absolute_path: bool = False,
    show_dl_link: bool = False,
    show_hash_link: bool = False,
    hash_link_protocol: str = PcsRapidUploadInfo.default_hash_link_protocol(),
    check_md5: bool = True,
    csv: bool = False,
    only_dl_link: bool = False,
    only_hash_link: bool = False,
):
    for rp in remotepaths:
        list_file(
            api,
            rp,
            desc=desc,
            name=name,
            time=time,
            size=size,
            recursive=recursive,
            sifters=sifters,
            highlight=highlight,
            rapiduploadinfo_file=rapiduploadinfo_file,
            user_id=user_id,
            user_name=user_name,
            show_size=show_size,
            show_date=show_date,
            show_md5=show_md5,
            show_absolute_path=show_absolute_path,
            show_dl_link=show_dl_link,
            show_hash_link=show_hash_link,
            hash_link_protocol=hash_link_protocol,
            check_md5=check_md5,
            csv=csv,
            only_dl_link=only_dl_link,
            only_hash_link=only_hash_link,
        )



import datetime
import time
from pathlib import Path
from typing import Any, Dict, List, Optional, Pattern, Tuple, Union

from ..pcs import (
    CloudTask,
    FromTo,
    PcsFile,
    PcsRapidUploadInfo,
    PcsSharedLink,
    PcsSharedPath,
    PcsUser,
)
from ..utils import format_date, format_time, human_size
from .sifter import Sifter

_print = print

from rich import print
from rich.box import SIMPLE
from rich.console import Console
from rich.highlighter import Highlighter as RichHighlighter
from rich.panel import Panel
from rich.style import Style
from rich.table import Table
from rich.text import Text


class Highlighter(RichHighlighter):
    def __init__(self, patterns: List[Union[Pattern, str]], style: Union[str, Style]):
        super().__init__()
        self.patterns = patterns
        self.style = style

    def highlight(self, text: Text):
        for pat in self.patterns:
            if isinstance(pat, Pattern):
                for m in pat.finditer(text.plain):
                    text.stylize(style=self.style, start=m.start(), end=m.end())
            else:
                pat_len = len(pat)
                start = 0
                while True:
                    idx = text.plain.find(pat, start)
                    if idx == -1:
                        break
                    text.stylize(style=self.style, start=idx, end=idx + pat_len)
                    start = idx + 1


def display_files(
    pcs_files: List[PcsFile],
    remotepath: Optional[str],
    sifters: List[Sifter] = [],
    highlight: bool = False,
    show_size: bool = False,
    show_date: bool = False,
    show_md5: bool = False,
    show_absolute_path: bool = False,
    show_dl_link: bool = False,
    show_hash_link: bool = False,
    hash_link_protocol: str = PcsRapidUploadInfo.default_hash_link_protocol(),
    csv: bool = False,
):
    if not pcs_files:
        return

    table = Table(box=SIMPLE, padding=0, show_edge=False)
    table.add_column()
    headers = []  # for csv
    headers.append("\t")
    if show_size:
        header = "Size"
        table.add_column(header, justify="right")
        headers.append(header)
    if show_date:
        header = "Modified Time"
        table.add_column(header, justify="center")
        headers.append(header)
    if show_md5:
        header = "md5"
        table.add_column(header, justify="left")
        headers.append(header)
    header = "Path"
    table.add_column(header, justify="left", overflow="fold")
    headers.append(header)
    if show_dl_link:
        header = "Download Link"
        table.add_column(header, justify="left", overflow="fold")
        headers.append(header)
    if show_hash_link:
        header = "Hash Link"
        table.add_column(header, justify="left", overflow="fold")
        headers.append(header)

    rows = []  # for csv

    max_size_str_len = max([len(str(pcs_file.size)) for pcs_file in pcs_files])
    for pcs_file in pcs_files:
        row: List[Union[str, Text]] = []

        if csv:
            row.append("-")
        else:
            tp = Text("-", style="bold red")
            row.append(tp)

        if show_size:
            size = human_size(pcs_file.size) if pcs_file.size else ""
            if csv:
                row.append(f"{size} {pcs_file.size}")
            else:
                row.append(f"{size} {pcs_file.size: >{max_size_str_len}}")
        if show_date:
            date = format_date(pcs_file.local_mtime) if pcs_file.local_mtime else ""
            row.append(date)
        if show_md5:
            md5 = pcs_file.md5 or ""
            row.append(md5)

        path = pcs_file.path if show_absolute_path else Path(pcs_file.path).name
        background = Text()
        if pcs_file.is_dir:
            if csv:
                row[0] = "d"
            else:
                tp._text = ["d"]
                background.style = "blue"

        if highlight and sifters:
            pats: List[Union[Pattern, str]] = list(
                filter(
                    None, [sifter.pattern() for sifter in sifters if sifter.include()]
                )
            )
            highlighter = Highlighter(pats, "yellow")
            _path = highlighter(path)
        else:
            _path = Text(path)

        if csv:
            row.append(path)
        else:
            row.append(background + _path)

        if show_dl_link:
            row.append(pcs_file.dl_link or "")

        rpinfo = pcs_file.rapid_upload_info
        if show_hash_link:
            link = ""
            if rpinfo:
                link = rpinfo.cs3l()
            row.append(link)

        if csv:
            rows.append(row)
        else:
            table.add_row(*row)

    if csv:
        _print(remotepath)
        _print("\t".join(headers))
        for row in rows:
            _print("\t".join(row))  # type: ignore
    else:
        console = Console()
        if remotepath:
            title = Text(remotepath, style="italic green")
            console.print(title)
        console.print(table)


def display_rapid_upload_links(
    infos: List[Dict[str, Any]],
    hash_link_protocol: str = PcsRapidUploadInfo.default_hash_link_protocol(),
    only_hash_link: bool = False,
):
    rpinfos = [
        (
            PcsRapidUploadInfo(
                slice_md5=r["slice_md5"],
                content_md5=r["content_md5"],
                content_crc32=r["content_crc32"],
                content_length=r["content_length"],
                remotepath=r["remotepath"],
            ),
            r["id"],
        )
        for r in infos
    ]

    if only_hash_link:
        for rpinfo, id in rpinfos:
            link = getattr(rpinfo, hash_link_protocol)()
            print(link)
        return

    _print("Id\tRemotepath\tLink")
    for rpinfo, id in rpinfos:
        row = [str(id), rpinfo.remotepath or ""]
        row.append(getattr(rpinfo, hash_link_protocol)())

        _print(*row, sep="\t")


def display_rapid_upload_infos(infos: List[Dict[str, Any]]):
    rpinfos = [
        PcsRapidUploadInfo(
            slice_md5=r["slice_md5"],
            content_md5=r["content_md5"],
            content_crc32=r["content_crc32"],
            content_length=r["content_length"],
            remotepath=r["remotepath"],
        )
        for r in infos
    ]

    panels = []
    for info, rpinfo in zip(infos, rpinfos):
        cn = "\n".join([f"{k}: {v}" for k, v in info.items()])
        cn += "\nHash Links:\n" + "".join(f"  {link}\n" for link in rpinfo.all_links())
        panel = Panel(cn, highlight=True)
        panels.append(panel)

    console = Console()
    console.print(*panels)


def display_from_to(*from_to_list: FromTo):
    if not from_to_list:
        return

    table = Table(box=SIMPLE, padding=0, show_edge=False)
    table.add_column("From", justify="left", overflow="fold")
    table.add_column("To", justify="left", overflow="fold")

    for from_to in from_to_list:
        table.add_row(from_to.from_, from_to.to_)

    console = Console()
    console.print(table)


_TASK_FORMAT = (
    "task_id: [bold white]{task_id}[/bold white]\n"
    "task name: [bold green]{task_name}[/bold green]\n"
    "source_url: [bold]{source_url}[/bold]\n"
    "remotepath: {path}\n"
    "status: [red]{status}[/red]\n"
    "[bold white]{percent}[/bold white]  "
    "[bold yellow]{finished_size}/{size}[/bold yellow]"
)


def display_tasks(*tasks: CloudTask):
    panels = []
    for task in tasks:
        status = task.status_mean()
        size = human_size(task.size)
        finished_size = human_size(task.finished_size)
        if task.size != 0:
            percent = f"{task.finished_size / task.size * 100:.1f}%"
        else:
            percent = "0%"

        info = task._asdict()
        info["status"] = status
        info["percent"] = percent
        info["finished_size"] = finished_size
        info["size"] = size

        panel = Panel(_TASK_FORMAT.format(**info), highlight=True)
        panels.append(panel)

    console = Console()
    console.print(*panels)


_SHARED_LINK_INFO_FORMAT = (
    "share id: {share_id}\n"
    "shared url: [bold]{url}[/bold]\n"
    "valid period: [i yellow]{expired_time}[/i yellow]\n"
    "password: [bold red]{password}[/bold red]\n"
    "paths: {paths}"
)


def display_shared_links(*shared_links: PcsSharedLink):
    panels = []
    for shared_link in shared_links:
        share_id = shared_link.share_id
        url = shared_link.url

        if shared_link.expired == -1:
            expired_time = "已经超期"
        elif shared_link.expired is None or shared_link.expired == 0:
            expired_time = "永久"
        else:
            expired_time = format_time(shared_link.expired or 0)

        password = shared_link.password or ""
        paths = "\n       ".join(shared_link.paths or [])

        panel = Panel(
            _SHARED_LINK_INFO_FORMAT.format(
                share_id=share_id,
                url=url,
                expired_time=expired_time,
                password=password,
                paths=paths,
            ),
            highlight=True,
        )
        panels.append(panel)

    console = Console()
    console.print(*panels)


def display_shared_paths(*shared_paths: PcsSharedPath):
    table = Table(box=SIMPLE, padding=0, show_edge=False)
    table.add_column()
    table.add_column("Size", justify="right")
    table.add_column("Path", justify="left", overflow="fold")

    max_size_str_len = max([len(str(shared_path.size)) for shared_path in shared_paths])
    for shared_path in shared_paths:
        row: List[Union[str, Text]] = []

        # Is file
        tp = Text("-", style="bold red")
        row.append(tp)

        size = human_size(shared_path.size) if shared_path.size else ""
        row.append(f"{size} {shared_path.size: >{max_size_str_len}}")

        path = shared_path.path
        background = Text()
        if shared_path.is_dir:
            tp._text = ["d"]
            background.style = "blue"

        _path = Text(path)
        row.append(background + _path)

        table.add_row(*row)

    console = Console()
    console.print(table)


def display_user_info(user_info: PcsUser):
    user_id, user_name, auth, age, sex, quota, products, level = user_info
    bduss = auth and auth.bduss
    quota_str = ""
    if quota:
        quota_str = human_size(quota.used) + "/" + human_size(quota.quota)

    products_str = ""
    for p in products or []:
        name = p.name
        start_date = format_date(p.start_time)
        end_date = format_date(p.end_time)
        avail = str(datetime.timedelta(seconds=(int(p.end_time - time.time()))))
        value = f"From {start_date} to {end_date}, left {avail}"
        products_str += f"\n    {name}: {value}"

    _tempt = (
        f"user id: {user_id}\n"
        f"user name: {user_name}\n"
        f"bduss: {bduss}\n"
        f"age: {age}\n"
        f"sex: {sex}\n"
        f"quota: {quota_str}\n"
        f"level: {level}\n"
        f"products:{products_str}\n"
    )

    console = Console()
    console.print(_tempt, highlight=True)


def display_user_infos(
    *user_infos: Tuple[PcsUser, str, str], recent_user_id: Optional[int] = None
):
    """
    Args:
        user_infos (*Tuple[PcsUser, pwd: str])
    """

    table = Table(box=SIMPLE, show_edge=False, highlight=True)
    table.add_column("Recent", justify="right")
    table.add_column("Index", justify="left")
    table.add_column("Account Name", justify="left", overflow="fold")
    table.add_column("User Id", justify="left", overflow="fold")
    table.add_column("Quota", justify="left")
    table.add_column("SVIP", justify="left", overflow="fold")
    table.add_column("VIP", justify="left", overflow="fold")
    table.add_column("Level", justify="left")
    table.add_column("pwd", justify="left", overflow="fold")

    for idx, (user_info, pwd, account_name) in enumerate(user_infos, 1):
        user_id, _, auth, age, sex, quota, products, level = user_info

        is_recent = "[green]✔[/green]" if user_id == recent_user_id else ""

        quota_str = ""
        if quota:
            quota_str = human_size(quota.used) + "/" + human_size(quota.quota)

        svip = "[red]✘[/red]"
        vip = "[red]✘[/red]"

        assert products
        for p in products or []:
            avail = str(datetime.timedelta(seconds=(int(p.end_time - time.time()))))
            if p.name.startswith("svip2_nd"):
                svip = f"Left [green]{avail}[/green]"
                continue
            if p.name.startswith("contentvip_nd"):
                vip = f"Left [green]{avail}[/green]"
                continue

        table.add_row(
            is_recent,
            str(idx),
            account_name,
            str(user_id),
            quota_str,
            svip,
            vip,
            str(level),
            pwd,
        )

    console = Console()
    console.print(table)


def display_blocked_remotepath(remotepath: str):
    print(f"[i yellow]Remote path is blocked[/i yellow]: {remotepath}")



from typing import Optional, List, Dict, NamedTuple
from collections import namedtuple
from urllib.parse import unquote

from base64 import standard_b64encode

import os


class PcsRapidUploadInfo(NamedTuple):
    """Rapid Upload Info"""

    slice_md5: str
    content_md5: str
    content_crc32: int
    content_length: int
    remotepath: Optional[str] = None

    def _filename(self) -> str:
        filename = os.path.basename(self.remotepath or "")
        if not filename:
            return ""
        else:
            return filename.replace(" ", "%20")

    def cs3l(self) -> str:
        """cs3l://<content_md5>#<slice_md5>#<content_crc32>#<content_length>#<filename>"""

        filename = self._filename()
        content_crc32 = self.content_crc32 or "0"
        return f"cs3l://{self.content_md5}#{self.slice_md5}#{content_crc32}#{self.content_length}#{filename}"

    def short(self) -> str:
        """<content_md5>#<slice_md5>#<content_length>#<filename>"""

        filename = self._filename()
        return f"{self.content_md5}#{self.slice_md5}#{self.content_length}#{filename}"

    def bdpan(self) -> str:
        """bdpan://{base64(<filename>|<content_length>|<content_md5>|<slice_md5>)}"""

        filename = self._filename()
        return "bdpan://" + standard_b64encode(
            f"{filename}|{self.content_length}|{self.content_md5}|{self.slice_md5}".encode(
                "utf-8"
            )
        ).decode("utf-8")

    def all_links(self) -> List[str]:
        return [self.cs3l(), self.short(), self.bdpan()]

    @staticmethod
    def hash_link_protocols() -> List[str]:
        return ["cs3l", "short", "bdpan"]

    @staticmethod
    def default_hash_link_protocol() -> str:
        return "cs3l"


class PcsFile(NamedTuple):
    """
    A Baidu PCS file

    path: str  # remote absolute path
    is_dir: Optional[bool] = None
    is_file: Optional[bool] = None
    fs_id: Optional[int] = None  # file id
    size: Optional[int] = None
    md5: Optional[str] = None
    block_list: Optional[List[str]] = None  # block md5 list
    category: Optional[int] = None
    user_id: Optional[int] = None
    ctime: Optional[int] = None  # server created time
    mtime: Optional[int] = None  # server modifed time
    local_ctime: Optional[int] = None  # local created time
    local_mtime: Optional[int] = None  # local modifed time
    server_ctime: Optional[int] = None  # server created time
    server_mtime: Optional[int] = None  # server modifed time
    shared: Optional[bool] = None  # this file is shared if True
    """

    path: str  # remote absolute path
    is_dir: Optional[bool] = None
    is_file: Optional[bool] = None
    fs_id: Optional[int] = None  # file id
    size: Optional[int] = None
    md5: Optional[str] = None
    block_list: Optional[List[str]] = None  # block md5 list
    category: Optional[int] = None
    user_id: Optional[int] = None
    ctime: Optional[int] = None  # server created time
    mtime: Optional[int] = None  # server modifed time
    local_ctime: Optional[int] = None  # local created time
    local_mtime: Optional[int] = None  # local modifed time
    server_ctime: Optional[int] = None  # server created time
    server_mtime: Optional[int] = None  # server modifed time
    shared: Optional[bool] = None  # this file is shared if True

    rapid_upload_info: Optional[PcsRapidUploadInfo] = None
    dl_link: Optional[str] = None

    @staticmethod
    def from_(info) -> "PcsFile":
        return PcsFile(
            path=info.get("path"),
            is_dir=info.get("isdir") == 1,
            is_file=info.get("isdir") == 0,
            fs_id=info.get("fs_id"),
            size=info.get("size"),
            md5=info.get("md5"),
            block_list=info.get("block_list"),
            category=info.get("category"),
            user_id=info.get("user_id"),
            ctime=info.get("ctime"),
            mtime=info.get("mtime"),
            local_ctime=info.get("local_ctime"),
            local_mtime=info.get("local_mtime"),
            server_ctime=info.get("server_ctime"),
            server_mtime=info.get("server_mtime"),
            shared=info.get("shared"),
        )


class PcsMagnetFile(NamedTuple):
    """
    A file in a magnet link

    path: str  # the file path in the magnet link
    size: Optional[int] = None
    """

    path: str  # the file path in the magnet link
    size: Optional[int] = None

    @staticmethod
    def from_(info) -> "PcsMagnetFile":
        return PcsMagnetFile(path=info.get("file_name"), size=info.get("size"))


class PcsSharedLink(NamedTuple):
    url: str
    paths: Optional[List[str]] = None
    fs_ids: Optional[List[int]] = None
    password: Optional[str] = None

    # The remained second before expiring.
    # -1 means being expired
    expired: Optional[int] = None

    # channel == 4, has password
    channel: Optional[bool] = None

    share_id: Optional[int] = None
    ctime: Optional[int] = None

    @staticmethod
    def from_(info) -> "PcsSharedLink":
        return PcsSharedLink(
            url=info.get("link") or info.get("shortlink"),
            paths=info.get("paths") or [info.get("typicalPath")],
            fs_ids=info.get("fsIds"),
            password=info.get("password"),
            expired=info.get("expiredTime") if info.get("expiredType") != -1 else -1,
            channel=info.get("channel"),
            share_id=info.get("share_id") or info.get("shareId") or info.get("shareid"),
            ctime=info.get("ctime"),
        )

    def has_password(self) -> bool:
        if self.password:
            return True
        if self.channel == 4:
            return True
        return False

    def available(self) -> bool:
        if not self.paths:
            return False
        if self.paths[0].startswith("/"):
            return True
        else:
            return False


class PcsSharedPath(NamedTuple):
    """
    User shared path

    `sharedpath`: original shared path
    `remotepath`: the directory where the `sharedpath` will save
    """

    fs_id: int
    path: str
    size: int
    is_dir: bool
    is_file: bool
    md5: Optional[str] = None
    local_ctime: Optional[int] = None  # local created time
    local_mtime: Optional[int] = None  # local modifed time
    server_ctime: Optional[int] = None  # server created time
    server_mtime: Optional[int] = None  # server modifed time

    uk: Optional[int] = None
    share_id: Optional[int] = None
    bdstoken: Optional[str] = None

    @staticmethod
    def from_(info) -> "PcsSharedPath":
        if "parent_path" in info:
            path = unquote(info["parent_path"] or "") + "/" + info["server_filename"]
        else:
            path = info["path"]
        return PcsSharedPath(
            fs_id=info.get("fs_id"),
            path=path,
            size=info.get("size"),
            is_dir=info.get("isdir") == 1,
            is_file=info.get("isdir") == 0,
            md5=info.get("md5"),
            local_ctime=info.get("local_ctime"),
            local_mtime=info.get("local_mtime"),
            server_ctime=info.get("server_ctime"),
            server_mtime=info.get("server_mtime"),
            uk=info.get("uk"),
            share_id=info.get("share_id") or info.get("shareid"),
            bdstoken=info.get("bdstoken"),
        )


FromTo = namedtuple("FromTo", ["from_", "to_"])


class PcsQuota(NamedTuple):
    quota: int
    used: int


class PcsAuth(NamedTuple):
    bduss: str
    cookies: Dict[str, Optional[str]]
    stoken: Optional[str] = None
    ptoken: Optional[str] = None


class PcsUserProduct(NamedTuple):
    name: str
    start_time: int  # second
    end_time: int  # second


class PcsUser(NamedTuple):
    user_id: int
    user_name: Optional[str] = None
    auth: Optional[PcsAuth] = None
    age: Optional[float] = None
    sex: Optional[str] = None
    quota: Optional[PcsQuota] = None
    products: Optional[List[PcsUserProduct]] = None
    level: Optional[int] = None


TASK_STATUS_MSG = {
    0: "下载成功",
    1: "下载进行中",
    2: "系统错误",
    3: "资源不存在",
    4: "下载超时",
    5: "资源存在但下载失败",
    6: "存储空间不足",
    7: "任务取消",
    8: "目标地址数据已存在",
    9: "任务删除",
    10: "部分下载成功",
}


class CloudTask(NamedTuple):
    """
    Baidu PCS Cloud Task
    """

    task_id: str
    source_url: str
    task_name: str
    path: str  # Saved remote directory
    status: int
    size: int
    finished_size: int
    ctime: int  # created time
    stime: int  # begined time
    ftime: int  # finished time

    @staticmethod
    def from_(info) -> "CloudTask":
        size = info.get("size") or info.get("file_size")
        if size:
            size = int(size)
        finished_size = info.get("finished_size")
        if finished_size:
            finished_size = int(finished_size)

        return CloudTask(
            task_id=str(info["task_id"]),
            source_url=info.get("source_url"),
            task_name=info.get("task_name"),
            path=info.get("save_path") or info.get("path"),
            status=int(info.get("status", 3)),
            size=size,
            finished_size=finished_size,
            ctime=info.get("ctime"),
            stime=info.get("stime"),
            ftime=info.get("ftime"),
        )

    def status_mean(self) -> Optional[str]:
        return TASK_STATUS_MSG.get(self.status)

    def finished(self) -> bool:
        return self.status == 0
